import inspect
import logging
import os
import socket
import subprocess
import sys
import textwrap
import threading
import time
import traceback

import salt.utils.files
import salt.utils.win_runas
import yaml
from tests.support.case import ModuleCase
from tests.support.helpers import with_system_user
from tests.support.mock import Mock
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import skipIf

try:
    import win32service
    import win32serviceutil
    import win32event
    import servicemanager
    import win32api

    CODE_DIR = win32api.GetLongPathName(RUNTIME_VARS.CODE_DIR)
    HAS_WIN32 = True
except ImportError:
    # Mock win32serviceutil object to avoid
    # a stacktrace in the _ServiceManager class
    win32serviceutil = Mock()
    HAS_WIN32 = False

logger = logging.getLogger(__name__)

PASSWORD = "P@ssW0rd"
NOPRIV_STDERR = "ERROR: Logged-on user does not have administrative privilege.\n"
PRIV_STDOUT = (
    "\nINFO: The system global flag 'maintain objects list' needs\n      "
    "to be enabled to see local opened files.\n      See Openfiles "
    "/? for more information.\n\n\nFiles opened remotely via local share "
    "points:\n---------------------------------------------\n\n"
    "INFO: No shared open files found.\n"
)
if HAS_WIN32:
    RUNAS_PATH = os.path.abspath(os.path.join(CODE_DIR, "runas.py"))
    RUNAS_OUT = os.path.abspath(os.path.join(CODE_DIR, "runas.out"))


def default_target(service, *args, **kwargs):
    while service.active:
        time.sleep(service.timeout)


class _ServiceManager(win32serviceutil.ServiceFramework):
    """
    A windows service manager
    """

    _svc_name_ = "Service Manager"
    _svc_display_name_ = "Service Manager"
    _svc_description_ = "A Service Manager"
    run_in_foreground = False
    target = default_target

    def __init__(self, args, target=None, timeout=60, active=True):
        win32serviceutil.ServiceFramework.__init__(self, args)
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)
        self.timeout = timeout
        self.active = active
        if target is not None:
            self.target = target

    @classmethod
    def log_error(cls, msg):
        if cls.run_in_foreground:
            logger.error(msg)
        servicemanager.LogErrorMsg(msg)

    @classmethod
    def log_info(cls, msg):
        if cls.run_in_foreground:
            logger.info(msg)
        servicemanager.LogInfoMsg(msg)

    @classmethod
    def log_exception(cls, msg):
        if cls.run_in_foreground:
            logger.exception(msg)
        exc_info = sys.exc_info()
        tb = traceback.format_tb(exc_info[2])
        servicemanager.LogErrorMsg("{} {} {}".format(msg, exc_info[1], tb))

    @property
    def timeout_ms(self):
        return self.timeout * 1000

    def SvcStop(self):
        """
        Stop the service by; terminating any subprocess call, notify
        windows internals of the stop event, set the instance's active
        attribute to 'False' so the run loops stop.
        """
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        win32event.SetEvent(self.hWaitStop)
        self.active = False

    def SvcDoRun(self):
        """
        Run the monitor in a separete thread so the main thread is
        free to react to events sent to the windows service.
        """
        servicemanager.LogMsg(
            servicemanager.EVENTLOG_INFORMATION_TYPE,
            servicemanager.PYS_SERVICE_STARTED,
            (self._svc_name_, ""),
        )
        self.log_info("Starting Service {}".format(self._svc_name_))
        monitor_thread = threading.Thread(target=self.target_thread)
        monitor_thread.start()
        while self.active:
            rc = win32event.WaitForSingleObject(self.hWaitStop, self.timeout_ms)
            if rc == win32event.WAIT_OBJECT_0:
                # Stop signal encountered
                self.log_info("Stopping Service")
                break
            if not monitor_thread.is_alive():
                self.log_info("Update Thread Died, Stopping Service")
                break

    def target_thread(self, *args, **kwargs):
        """
        Target Thread, handles any exception in the target method and
        logs them.
        """
        self.log_info("Monitor")
        try:
            self.target(self, *args, **kwargs)
        except Exception as exc:  # pylint: disable=broad-except
            # TODO: Add traceback info to windows event log objects
            self.log_exception("Exception In Target")

    @classmethod
    def install(cls, username=None, password=None, start_type=None):
        if hasattr(cls, "_svc_reg_class_"):
            svc_class = cls._svc_reg_class_
        else:
            svc_class = win32serviceutil.GetServiceClassString(cls)
        win32serviceutil.InstallService(
            svc_class,
            cls._svc_name_,
            cls._svc_display_name_,
            description=cls._svc_description_,
            userName=username,
            password=password,
            startType=start_type,
        )

    @classmethod
    def remove(cls):
        win32serviceutil.RemoveService(cls._svc_name_)

    @classmethod
    def start(cls):
        win32serviceutil.StartService(cls._svc_name_)

    @classmethod
    def restart(cls):
        win32serviceutil.RestartService(cls._svc_name_)

    @classmethod
    def stop(cls):
        win32serviceutil.StopService(cls._svc_name_)


def service_class_factory(
    cls_name,
    name,
    target=default_target,
    display_name="",
    description="",
    run_in_foreground=False,
):
    frm = inspect.stack()[1]
    mod = inspect.getmodule(frm[0])
    return type(
        cls_name,
        (_ServiceManager, object),
        {
            "__module__": mod.__name__,
            "_svc_name_": name,
            "_svc_display_name_": display_name or name,
            "_svc_description_": description,
            "run_in_foreground": run_in_foreground,
            "target": target,
        },
    )


if HAS_WIN32:
    test_service = service_class_factory("test_service", "test service")


SERVICE_SOURCE = """
from __future__ import absolute_import, unicode_literals
import logging
logger = logging.getLogger()
logging.basicConfig(level=logging.DEBUG, format="%(message)s")

from tests.integration.utils.test_win_runas import service_class_factory
import salt.utils.win_runas
import sys
import yaml

OUTPUT = {}
USERNAME = '{}'
PASSWORD = '{}'


def target(service, *args, **kwargs):
    service.log_info("target start")
    if PASSWORD:
        ret = salt.utils.win_runas.runas(
            'cmd.exe /C OPENFILES',
            username=USERNAME,
            password=PASSWORD,
        )
    else:
        ret = salt.utils.win_runas.runas(
            'cmd.exe /C OPENFILES',
            username=USERNAME,
        )

    service.log_info("win_runas returned %s" % ret)
    with salt.utils.files.fopen(OUTPUT, 'w') as fp:
        yaml.dump(ret, fp)
    service.log_info("target stop")


# This class will get imported and run as the service
test_service = service_class_factory('test_service', 'test service', target=target)

if __name__ == '__main__':
    try:
        test_service.stop()
    except Exception as exc:  # pylint: disable=broad-except
        logger.debug("stop service failed, this is ok.")
    try:
        test_service.remove()
    except Exception as exc:  # pylint: disable=broad-except
        logger.debug("remove service failed, this os ok.")
    test_service.install()
    sys.exit(0)
"""


def wait_for_service(name, timeout=200):
    start = time.time()
    while True:
        status = win32serviceutil.QueryServiceStatus(name)
        if status[1] == win32service.SERVICE_STOPPED:
            break
        if time.time() - start > timeout:
            raise TimeoutError(
                "Timeout waiting for service"
            )  # pylint: disable=undefined-variable

        time.sleep(0.3)


@skipIf(not HAS_WIN32, "This test runs only on windows.")
class RunAsTest(ModuleCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.hostname = socket.gethostname()

    @with_system_user(
        "test-runas", on_existing="delete", delete=True, password=PASSWORD
    )
    def test_runas(self, username):
        ret = salt.utils.win_runas.runas("cmd.exe /C OPENFILES", username, PASSWORD)
        self.assertEqual(ret["stdout"], "")
        self.assertEqual(ret["stderr"], NOPRIV_STDERR)
        self.assertEqual(ret["retcode"], 1)

    @with_system_user(
        "test-runas", on_existing="delete", delete=True, password=PASSWORD
    )
    def test_runas_no_pass(self, username):
        ret = salt.utils.win_runas.runas("cmd.exe /C OPENFILES", username)
        self.assertEqual(ret["stdout"], "")
        self.assertEqual(ret["stderr"], NOPRIV_STDERR)
        self.assertEqual(ret["retcode"], 1)

    @with_system_user(
        "test-runas-admin",
        on_existing="delete",
        delete=True,
        password=PASSWORD,
        groups=["Administrators"],
    )
    def test_runas_admin(self, username):
        ret = salt.utils.win_runas.runas("cmd.exe /C OPENFILES", username, PASSWORD)
        self.assertEqual(ret["stdout"], PRIV_STDOUT)
        self.assertEqual(ret["stderr"], "")
        self.assertEqual(ret["retcode"], 0)

    @with_system_user(
        "test-runas-admin",
        on_existing="delete",
        delete=True,
        password=PASSWORD,
        groups=["Administrators"],
    )
    def test_runas_admin_no_pass(self, username):
        ret = salt.utils.win_runas.runas("cmd.exe /C OPENFILES", username)
        self.assertEqual(ret["stdout"], PRIV_STDOUT)
        self.assertEqual(ret["stderr"], "")
        self.assertEqual(ret["retcode"], 0)

    def test_runas_system_user(self):
        ret = salt.utils.win_runas.runas("cmd.exe /C OPENFILES", "SYSTEM")
        self.assertEqual(ret["stdout"], PRIV_STDOUT)
        self.assertEqual(ret["stderr"], "")
        self.assertEqual(ret["retcode"], 0)

    def test_runas_network_service(self):
        ret = salt.utils.win_runas.runas("cmd.exe /C OPENFILES", "NETWORK SERVICE")
        self.assertEqual(ret["stdout"], "")
        self.assertEqual(ret["stderr"], NOPRIV_STDERR)
        self.assertEqual(ret["retcode"], 1)

    def test_runas_local_service(self):
        ret = salt.utils.win_runas.runas("cmd.exe /C OPENFILES", "LOCAL SERVICE")
        self.assertEqual(ret["stdout"], "")
        self.assertEqual(ret["stderr"], NOPRIV_STDERR)
        self.assertEqual(ret["retcode"], 1)

    @with_system_user(
        "test-runas", on_existing="delete", delete=True, password=PASSWORD
    )
    def test_runas_winrs(self, username):
        runaspy = textwrap.dedent(
            """
        import sys
        import salt.utils.win_runas
        username = '{}'
        password = '{}'
        sys.exit(salt.utils.win_runas.runas('cmd.exe /C OPENFILES', username, password)['retcode'])
        """.format(
                username, PASSWORD
            )
        )
        with salt.utils.files.fopen(RUNAS_PATH, "w") as fp:
            fp.write(runaspy)
        ret = subprocess.call(
            [
                "cmd.exe",
                "/C",
                "winrs",
                "/r:{}".format(self.hostname),
                "python",
                RUNAS_PATH,
            ]
        )
        self.assertEqual(ret, 1)

    @with_system_user(
        "test-runas", on_existing="delete", delete=True, password=PASSWORD
    )
    def test_runas_winrs_no_pass(self, username):
        runaspy = textwrap.dedent(
            """
        import sys
        import salt.utils.win_runas
        username = '{}'
        sys.exit(salt.utils.win_runas.runas('cmd.exe /C OPENFILES', username)['retcode'])
        """.format(
                username
            )
        )
        with salt.utils.files.fopen(RUNAS_PATH, "w") as fp:
            fp.write(runaspy)
        ret = subprocess.call(
            [
                "cmd.exe",
                "/C",
                "winrs",
                "/r:{}".format(self.hostname),
                "python",
                RUNAS_PATH,
            ]
        )
        self.assertEqual(ret, 1)

    @with_system_user(
        "test-runas-admin",
        on_existing="delete",
        delete=True,
        password=PASSWORD,
        groups=["Administrators"],
    )
    def test_runas_winrs_admin(self, username):
        runaspy = textwrap.dedent(
            """
        import sys
        import salt.utils.win_runas
        username = '{}'
        password = '{}'
        sys.exit(salt.utils.win_runas.runas('cmd.exe /C OPENFILES', username, password)['retcode'])
        """.format(
                username, PASSWORD
            )
        )
        with salt.utils.files.fopen(RUNAS_PATH, "w") as fp:
            fp.write(runaspy)
        ret = subprocess.call(
            [
                "cmd.exe",
                "/C",
                "winrs",
                "/r:{}".format(self.hostname),
                "python",
                RUNAS_PATH,
            ]
        )
        self.assertEqual(ret, 0)

    @with_system_user(
        "test-runas-admin",
        on_existing="delete",
        delete=True,
        password=PASSWORD,
        groups=["Administrators"],
    )
    def test_runas_winrs_admin_no_pass(self, username):
        runaspy = textwrap.dedent(
            """
        import sys
        import salt.utils.win_runas
        username = '{}'
        sys.exit(salt.utils.win_runas.runas('cmd.exe /C OPENFILES', username)['retcode'])
        """.format(
                username
            )
        )
        with salt.utils.files.fopen(RUNAS_PATH, "w") as fp:
            fp.write(runaspy)
        ret = subprocess.call(
            [
                "cmd.exe",
                "/C",
                "winrs",
                "/r:{}".format(self.hostname),
                "python",
                RUNAS_PATH,
            ]
        )
        self.assertEqual(ret, 0)

    def test_runas_winrs_system_user(self):
        runaspy = textwrap.dedent(
            """
        import sys
        import salt.utils.win_runas
        sys.exit(salt.utils.win_runas.runas('cmd.exe /C OPENFILES', 'SYSTEM')['retcode'])
        """
        )
        with salt.utils.files.fopen(RUNAS_PATH, "w") as fp:
            fp.write(runaspy)
        ret = subprocess.call(
            [
                "cmd.exe",
                "/C",
                "winrs",
                "/r:{}".format(self.hostname),
                "python",
                RUNAS_PATH,
            ]
        )
        self.assertEqual(ret, 0)

    def test_runas_winrs_network_service_user(self):
        runaspy = textwrap.dedent(
            """
        import sys
        import salt.utils.win_runas
        sys.exit(salt.utils.win_runas.runas('cmd.exe /C OPENFILES', 'NETWORK SERVICE')['retcode'])
        """
        )
        with salt.utils.files.fopen(RUNAS_PATH, "w") as fp:
            fp.write(runaspy)
        ret = subprocess.call(
            [
                "cmd.exe",
                "/C",
                "winrs",
                "/r:{}".format(self.hostname),
                "python",
                RUNAS_PATH,
            ]
        )
        self.assertEqual(ret, 1)

    def test_runas_winrs_local_service_user(self):
        runaspy = textwrap.dedent(
            """
        import sys
        import salt.utils.win_runas
        sys.exit(salt.utils.win_runas.runas('cmd.exe /C OPENFILES', 'LOCAL SERVICE')['retcode'])
        """
        )
        with salt.utils.files.fopen(RUNAS_PATH, "w") as fp:
            fp.write(runaspy)
        ret = subprocess.call(
            [
                "cmd.exe",
                "/C",
                "winrs",
                "/r:{}".format(self.hostname),
                "python",
                RUNAS_PATH,
            ]
        )
        self.assertEqual(ret, 1)

    @with_system_user(
        "test-runas", on_existing="delete", delete=True, password=PASSWORD
    )
    def test_runas_powershell_remoting(self, username):
        runaspy = textwrap.dedent(
            """
        import sys
        import salt.utils.win_runas
        username = '{}'
        password = '{}'
        sys.exit(salt.utils.win_runas.runas('cmd.exe /C OPENFILES', username, password)['retcode'])
        """.format(
                username, PASSWORD
            )
        )
        with salt.utils.files.fopen(RUNAS_PATH, "w") as fp:
            fp.write(runaspy)
        ret = subprocess.call(
            [
                "powershell",
                "Invoke-Command",
                "-ComputerName",
                self.hostname,
                "-ScriptBlock",
                "{{ python.exe {} }}".format(RUNAS_PATH),
            ]
        )
        self.assertEqual(ret, 1)

    @with_system_user(
        "test-runas", on_existing="delete", delete=True, password=PASSWORD
    )
    def test_runas_powershell_remoting_no_pass(self, username):
        runaspy = textwrap.dedent(
            """
        import sys
        import salt.utils.win_runas
        username = '{}'
        sys.exit(salt.utils.win_runas.runas('cmd.exe /C OPENFILES', username)['retcode'])
        """.format(
                username
            )
        )
        with salt.utils.files.fopen(RUNAS_PATH, "w") as fp:
            fp.write(runaspy)
        ret = subprocess.call(
            [
                "powershell",
                "Invoke-Command",
                "-ComputerName",
                self.hostname,
                "-ScriptBlock",
                "{{ python.exe {} }}".format(RUNAS_PATH),
            ]
        )
        self.assertEqual(ret, 1)

    @with_system_user(
        "test-runas-admin",
        on_existing="delete",
        delete=True,
        password=PASSWORD,
        groups=["Administrators"],
    )
    def test_runas_powershell_remoting_admin(self, username):
        psrp_wrap = (
            "powershell Invoke-Command -ComputerName {} -ScriptBlock {{ {} }}; exit"
            " $LASTEXITCODE"
        )
        runaspy = textwrap.dedent(
            """
        import sys
        import salt.utils.win_runas
        username = '{}'
        password = '{}'
        ret = salt.utils.win_runas.runas('cmd.exe /C OPENFILES', username, password)
        sys.exit(ret['retcode'])
        """.format(
                username, PASSWORD
            )
        )
        with salt.utils.files.fopen(RUNAS_PATH, "w") as fp:
            fp.write(runaspy)
        cmd = "python.exe {}; exit $LASTEXITCODE".format(RUNAS_PATH)
        ret = subprocess.call(psrp_wrap.format(self.hostname, cmd), shell=True)  # nosec
        self.assertEqual(ret, 0)

    @with_system_user(
        "test-runas-admin",
        on_existing="delete",
        delete=True,
        password=PASSWORD,
        groups=["Administrators"],
    )
    def test_runas_powershell_remoting_admin_no_pass(self, username):
        psrp_wrap = (
            "powershell Invoke-Command -ComputerName {} -ScriptBlock {{ {} }}; exit"
            " $LASTEXITCODE"
        )
        runaspy = textwrap.dedent(
            """
        import sys
        import salt.utils.win_runas
        username = '{}'
        sys.exit(salt.utils.win_runas.runas('cmd.exe /C OPENFILES', username)['retcode'])
        """.format(
                username
            )
        )
        with salt.utils.files.fopen(RUNAS_PATH, "w") as fp:
            fp.write(runaspy)
        cmd = "python.exe {}; exit $LASTEXITCODE".format(RUNAS_PATH)
        ret = subprocess.call(psrp_wrap.format(self.hostname, cmd), shell=True)  # nosec
        self.assertEqual(ret, 0)

    @with_system_user(
        "test-runas", on_existing="delete", delete=True, password=PASSWORD
    )
    def test_runas_service(self, username, timeout=200):
        if os.path.exists(RUNAS_OUT):
            os.remove(RUNAS_OUT)
        assert not os.path.exists(RUNAS_OUT)
        runaspy = SERVICE_SOURCE.format(repr(RUNAS_OUT), username, PASSWORD)
        with salt.utils.files.fopen(RUNAS_PATH, "w", encoding="utf-8") as fp:
            fp.write(runaspy)
        ret = subprocess.call(["python.exe", RUNAS_PATH])
        self.assertEqual(ret, 0)
        win32serviceutil.StartService("test service")
        wait_for_service("test service")
        with salt.utils.files.fopen(RUNAS_OUT, "r") as fp:
            ret = yaml.load(fp)
        assert ret["retcode"] == 1, ret

    @with_system_user(
        "test-runas", on_existing="delete", delete=True, password=PASSWORD
    )
    def test_runas_service_no_pass(self, username, timeout=200):
        if os.path.exists(RUNAS_OUT):
            os.remove(RUNAS_OUT)
        assert not os.path.exists(RUNAS_OUT)
        runaspy = SERVICE_SOURCE.format(repr(RUNAS_OUT), username, "")
        with salt.utils.files.fopen(RUNAS_PATH, "w", encoding="utf-8") as fp:
            fp.write(runaspy)
        ret = subprocess.call(["python.exe", RUNAS_PATH])
        self.assertEqual(ret, 0)
        win32serviceutil.StartService("test service")
        wait_for_service("test service")
        with salt.utils.files.fopen(RUNAS_OUT, "r") as fp:
            ret = yaml.load(fp)
        assert ret["retcode"] == 1, ret

    @with_system_user(
        "test-runas-admin",
        on_existing="delete",
        delete=True,
        password=PASSWORD,
        groups=["Administrators"],
    )
    def test_runas_service_admin(self, username, timeout=200):
        if os.path.exists(RUNAS_OUT):
            os.remove(RUNAS_OUT)
        assert not os.path.exists(RUNAS_OUT)
        runaspy = SERVICE_SOURCE.format(repr(RUNAS_OUT), username, PASSWORD)
        with salt.utils.files.fopen(RUNAS_PATH, "w", encoding="utf-8") as fp:
            fp.write(runaspy)
        ret = subprocess.call(["python.exe", RUNAS_PATH])
        self.assertEqual(ret, 0)
        win32serviceutil.StartService("test service")
        wait_for_service("test service")
        with salt.utils.files.fopen(RUNAS_OUT, "r") as fp:
            ret = yaml.load(fp)
        assert ret["retcode"] == 0, ret

    @with_system_user(
        "test-runas-admin",
        on_existing="delete",
        delete=True,
        password=PASSWORD,
        groups=["Administrators"],
    )
    def test_runas_service_admin_no_pass(self, username, timeout=200):
        if os.path.exists(RUNAS_OUT):
            os.remove(RUNAS_OUT)
        assert not os.path.exists(RUNAS_OUT)
        runaspy = SERVICE_SOURCE.format(repr(RUNAS_OUT), username, "")
        with salt.utils.files.fopen(RUNAS_PATH, "w", encoding="utf-8") as fp:
            fp.write(runaspy)
        ret = subprocess.call(["python.exe", RUNAS_PATH])
        self.assertEqual(ret, 0)
        win32serviceutil.StartService("test service")
        wait_for_service("test service")
        with salt.utils.files.fopen(RUNAS_OUT, "r") as fp:
            ret = yaml.load(fp)
        assert ret["retcode"] == 0, ret

    def test_runas_service_system_user(self):
        if os.path.exists(RUNAS_OUT):
            os.remove(RUNAS_OUT)
        assert not os.path.exists(RUNAS_OUT)
        runaspy = SERVICE_SOURCE.format(repr(RUNAS_OUT), "SYSTEM", "")
        with salt.utils.files.fopen(RUNAS_PATH, "w", encoding="utf-8") as fp:
            fp.write(runaspy)
        ret = subprocess.call(["python.exe", RUNAS_PATH])
        self.assertEqual(ret, 0)
        win32serviceutil.StartService("test service")
        wait_for_service("test service")
        with salt.utils.files.fopen(RUNAS_OUT, "r") as fp:
            ret = yaml.load(fp)
        assert ret["retcode"] == 0, ret



import os

import pytest
import salt.modules.cmdmod as cmd
import salt.modules.virtualenv_mod
import salt.modules.zcbuildout as modbuildout
import salt.states.zcbuildout as buildout
import salt.utils.path
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import skipIf
from tests.unit.modules.test_zcbuildout import KNOWN_VIRTUALENV_BINARY_NAMES, Base


@skipIf(
    salt.utils.path.which_bin(KNOWN_VIRTUALENV_BINARY_NAMES) is None,
    "The 'virtualenv' packaged needs to be installed",
)
@pytest.mark.requires_network
class BuildoutTestCase(Base):
    def setup_loader_modules(self):
        module_globals = {
            "__env__": "base",
            "__opts__": {"test": False},
            "__salt__": {
                "cmd.run_all": cmd.run_all,
                "cmd.run": cmd.run,
                "cmd.retcode": cmd.retcode,
                "buildout.buildout": modbuildout.buildout,
            },
        }
        return {buildout: module_globals, modbuildout: module_globals}

    # I don't have the time to invest in learning more about buildout,
    # and given we don't have support yet, and there are other priorities
    # I'm going to punt on this for now - WW
    @skipIf(True, "Buildout is still in beta. Test needs fixing.")
    def test_quiet(self):
        c_dir = os.path.join(self.tdir, "c")
        assert False, os.listdir(self.rdir)
        modbuildout.upgrade_bootstrap(c_dir)
        cret = buildout.installed(c_dir, python=self.py_st)
        self.assertFalse("OUTPUT:" in cret["comment"], cret["comment"])
        self.assertFalse("Log summary:" in cret["comment"], cret["comment"])
        self.assertTrue(cret["result"], cret["comment"])

    @pytest.mark.slow_test
    def test_error(self):
        b_dir = os.path.join(self.tdir, "e")
        ret = buildout.installed(b_dir, python=self.py_st)
        self.assertTrue("Unexpected response from buildout" in ret["comment"])
        self.assertFalse(ret["result"])

    @pytest.mark.slow_test
    def test_installed(self):
        if salt.modules.virtualenv_mod.virtualenv_ver(self.ppy_st) >= (20, 0, 0):
            self.skipTest(
                "Skiping until upstream resolved"
                " https://github.com/pypa/virtualenv/issues/1715"
            )
        b_dir = os.path.join(self.tdir, "b")
        ret = buildout.installed(
            b_dir, python=self.py_st, onlyif=RUNTIME_VARS.SHELL_FALSE_PATH
        )
        self.assertEqual(ret["comment"], "\nonlyif condition is false")
        self.assertEqual(ret["result"], True)
        self.assertTrue(os.sep + "b" in ret["name"])
        b_dir = os.path.join(self.tdir, "b")
        ret = buildout.installed(
            b_dir, python=self.py_st, unless=RUNTIME_VARS.SHELL_TRUE_PATH
        )
        self.assertEqual(ret["comment"], "\nunless condition is true")
        self.assertEqual(ret["result"], True)
        self.assertTrue(os.sep + "b" in ret["name"])
        ret = buildout.installed(b_dir, python=self.py_st)
        self.assertEqual(ret["result"], True)
        self.assertTrue("OUTPUT:" in ret["comment"])
        self.assertTrue("Log summary:" in ret["comment"])



"""
Load up the libvirt keys into Pillar for a given minion if said keys have been
generated using the libvirt key runner

:depends: certtool
"""
import os
import subprocess

import salt.utils.files
import salt.utils.path
import salt.utils.stringutils

# Don't "fix" the above docstring to put it on two lines, as the sphinx
# autosummary pulls only the first line for its description.


def __virtual__():
    return salt.utils.path.which("certtool") is not None


def ext_pillar(minion_id, pillar, command):
    """
    Read in the generated libvirt keys
    """
    key_dir = os.path.join(__opts__["pki_dir"], "libvirt", minion_id)
    cacert = os.path.join(__opts__["pki_dir"], "libvirt", "cacert.pem")
    if not os.path.isdir(key_dir):
        # No keys have been generated
        gen_hyper_keys(
            minion_id,
            pillar.get("ext_pillar_virt.country", "US"),
            pillar.get("ext_pillar_virt.st", "Utah"),
            pillar.get("ext_pillar_virt.locality", "Salt Lake City"),
            pillar.get("ext_pillar_virt.organization", "Salted"),
            pillar.get("ext_pillar_virt.expiration_days", "365"),
        )
    ret = {}
    for key in os.listdir(key_dir):
        if not key.endswith(".pem"):
            continue
        fn_ = os.path.join(key_dir, key)
        with salt.utils.files.fopen(fn_, "r") as fp_:
            ret["libvirt.{}".format(key)] = salt.utils.stringutils.to_unicode(
                fp_.read()
            )
    with salt.utils.files.fopen(cacert, "r") as fp_:
        ret["libvirt.cacert.pem"] = salt.utils.stringutils.to_unicode(fp_.read())
    return ret


def gen_hyper_keys(
    minion_id,
    country="US",
    state="Utah",
    locality="Salt Lake City",
    organization="Salted",
    expiration_days="365",
):
    """
    Generate the keys to be used by libvirt hypervisors, this routine gens
    the keys and applies them to the pillar for the hypervisor minions
    """
    key_dir = os.path.join(__opts__["pki_dir"], "libvirt")
    if not os.path.isdir(key_dir):
        os.makedirs(key_dir)
    cakey = os.path.join(key_dir, "cakey.pem")
    cacert = os.path.join(key_dir, "cacert.pem")
    cainfo = os.path.join(key_dir, "ca.info")
    if not os.path.isfile(cainfo):
        with salt.utils.files.fopen(cainfo, "w+") as fp_:
            fp_.write("cn = salted\nca\ncert_signing_key")
    if not os.path.isfile(cakey):
        proc = subprocess.run(
            ["certtool", "--generate-privkey"],
            stdout=subprocess.PIPE,
            universal_newlines=True,
            check=True,
        )
        with salt.utils.files.fopen(cakey, "w") as wfh:
            wfh.write(proc.stdout)
    if not os.path.isfile(cacert):
        subprocess.call(
            [
                "certtool",
                "--generate-self-signed",
                "--load-privkey",
                cakey,
                "--template",
                cainfo,
                "--outfile",
                cacert,
            ]
        )
    sub_dir = os.path.join(key_dir, minion_id)
    if not os.path.isdir(sub_dir):
        os.makedirs(sub_dir)
    priv = os.path.join(sub_dir, "serverkey.pem")
    cert = os.path.join(sub_dir, "servercert.pem")
    srvinfo = os.path.join(sub_dir, "server.info")
    cpriv = os.path.join(sub_dir, "clientkey.pem")
    ccert = os.path.join(sub_dir, "clientcert.pem")
    clientinfo = os.path.join(sub_dir, "client.info")
    if not os.path.isfile(srvinfo):
        with salt.utils.files.fopen(srvinfo, "w+") as fp_:
            infodat = salt.utils.stringutils.to_str(
                "organization = salted\ncn = {}\ntls_www_server"
                "\nencryption_key\nsigning_key"
                "\ndigitalSignature\nexpiration_days = {}".format(
                    __grains__["fqdn"], expiration_days
                )
            )
            fp_.write(infodat)
    if not os.path.isfile(priv):
        proc = subprocess.run(
            ["certtool", "--generate-privkey"],
            stdout=subprocess.PIPE,
            universal_newlines=True,
            check=True,
        )
        with salt.utils.files.fopen(priv, "w") as wfh:
            wfh.write(proc.stdout)
    if not os.path.isfile(cert):
        subprocess.call(
            [
                "certtool",
                "--generate-certificate",
                "--load-privkey",
                priv,
                "--load-ca-certificate",
                cacert,
                "--load-ca-privkey",
                cakey,
                "--template",
                srvinfo,
                "--outfile",
                cert,
            ]
        )
    if not os.path.isfile(clientinfo):
        with salt.utils.files.fopen(clientinfo, "w+") as fp_:
            infodat = salt.utils.stringutils.to_str(
                "country = {}\nstate = {}\nlocality = {}\n"
                "organization = {}\ncn = {}\n"
                "tls_www_client\nencryption_key\nsigning_key\n"
                "digitalSignature".format(
                    country, state, locality, organization, __grains__["fqdn"]
                )
            )
            fp_.write(infodat)
    if not os.path.isfile(cpriv):
        proc = subprocess.run(
            ["certtool", "--generate-privkey"],
            stdout=subprocess.PIPE,
            universal_newlines=True,
            check=True,
        )
        with salt.utils.files.fopen(cpriv, "w") as wfh:
            wfh.write(proc.stdout)
    if not os.path.isfile(ccert):
        subprocess.call(
            [
                "certtool",
                "--generate-certificate",
                "--load-privkey",
                cpriv,
                "--load-ca-certificate",
                cacert,
                "--load-ca-privkey",
                cakey,
                "--template",
                clientinfo,
                "--outfile",
                ccert,
            ]
        )



import pytest
import salt.modules.kubeadm as kubeadm
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class KubeAdmTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.kubeadm
    """

    def setup_loader_modules(self):
        return {kubeadm: {"__salt__": {}, "__utils__": {}}}

    def test_version(self):
        """
        Test kuebadm.version without parameters
        """
        version = '{"clientVersion":{"major":"1"}}'
        salt_mock = {
            "cmd.run_stdout": MagicMock(return_value=version),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.version() == {"clientVersion": {"major": "1"}}
            salt_mock["cmd.run_stdout"].assert_called_with(
                ["kubeadm", "version", "--output", "json"]
            )

    def test_version_params(self):
        """
        Test kuebadm.version with parameters
        """
        version = '{"clientVersion":{"major":"1"}}'
        salt_mock = {
            "cmd.run_stdout": MagicMock(return_value=version),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.version(kubeconfig="/kube.cfg", rootfs="/mnt") == {
                "clientVersion": {"major": "1"}
            }
            salt_mock["cmd.run_stdout"].assert_called_with(
                [
                    "kubeadm",
                    "version",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                    "--output",
                    "json",
                ]
            )

    def test_token_create(self):
        """
        Test kuebadm.token_create without parameters
        """
        result = {"retcode": 0, "stdout": "token"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.token_create() == "token"
            salt_mock["cmd.run_all"].assert_called_with(["kubeadm", "token", "create"])

    def test_token_create_params(self):
        """
        Test kuebadm.token_create with parameters
        """
        result = {"retcode": 0, "stdout": "token"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.token_create(
                    token="token",
                    config="/kubeadm.cfg",
                    description="a description",
                    groups=["g:1", "g:2"],
                    ttl="1h1m1s",
                    usages=["u1", "u2"],
                    kubeconfig="/kube.cfg",
                    rootfs="/mnt",
                )
                == "token"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "token",
                    "create",
                    "token",
                    "--config",
                    "/kubeadm.cfg",
                    "--description",
                    "a description",
                    "--groups",
                    '["g:1", "g:2"]',
                    "--ttl",
                    "1h1m1s",
                    "--usages",
                    '["u1", "u2"]',
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_token_create_error(self):
        """
        Test kuebadm.token_create error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.token_create()

    def test_token_delete(self):
        """
        Test kuebadm.token_delete without parameters
        """
        result = {"retcode": 0, "stdout": "deleted"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.token_delete("token")
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "token", "delete", "token"]
            )

    def test_token_delete_params(self):
        """
        Test kuebadm.token_delete with parameters
        """
        result = {"retcode": 0, "stdout": "deleted"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.token_delete("token", kubeconfig="/kube.cfg", rootfs="/mnt")
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "token",
                    "delete",
                    "token",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_token_delete_error(self):
        """
        Test kuebadm.token_delete error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.token_delete("token")

    def test_token_generate(self):
        """
        Test kuebadm.token_generate without parameters
        """
        result = {"retcode": 0, "stdout": "token"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.token_generate() == "token"
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "token", "generate"]
            )

    def test_token_generate_params(self):
        """
        Test kuebadm.token_generate with parameters
        """
        result = {"retcode": 0, "stdout": "token"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.token_generate(kubeconfig="/kube.cfg", rootfs="/mnt") == "token"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "token",
                    "generate",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_token_generate_error(self):
        """
        Test kuebadm.token_generate error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.token_generate()

    def test_token_empty(self):
        """
        Test kuebadm.token_list when no outout
        """
        result = {"retcode": 0, "stdout": ""}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.token_list() == []
            salt_mock["cmd.run_all"].assert_called_with(["kubeadm", "token", "list"])

    def test_token_list(self):
        """
        Test kuebadm.token_list without parameters
        """
        output = "H1  H2  H31 H32  H4\n1   2   3.1 3.2  4"
        result = {"retcode": 0, "stdout": output}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.token_list() == [
                {"h1": "1", "h2": "2", "h31 h32": "3.1 3.2", "h4": "4"}
            ]
            salt_mock["cmd.run_all"].assert_called_with(["kubeadm", "token", "list"])

    def test_token_list_multiple_lines(self):
        """
        Test kuebadm.token_list with multiple tokens
        """
        output = "H1  H2  H31 H32  H4\n1   2   3.1 3.2  4\na   b   c d      e"
        result = {"retcode": 0, "stdout": output}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.token_list() == [
                {"h1": "1", "h2": "2", "h31 h32": "3.1 3.2", "h4": "4"},
                {"h1": "a", "h2": "b", "h31 h32": "c d", "h4": "e"},
            ]

    def test_token_list_broken_lines(self):
        """
        Test kuebadm.token_list with multiple tokens, one broken
        """
        output = "H1  H2  H31 H32  H4\n1   2   3.1 3.2  4\na   b   c  d     e"
        result = {"retcode": 0, "stdout": output}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.token_list() == [
                {"h1": "1", "h2": "2", "h31 h32": "3.1 3.2", "h4": "4"}
            ]

    def test_token_list_params(self):
        """
        Test kuebadm.token_list with parameters
        """
        output = "H1  H2  H31 H32  H4\n1   2   3.1 3.2  4"
        result = {"retcode": 0, "stdout": output}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            result = kubeadm.token_list(kubeconfig="/kube.cfg", rootfs="/mnt")
            assert result == [{"h1": "1", "h2": "2", "h31 h32": "3.1 3.2", "h4": "4"}]
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "token",
                    "list",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_token_list_error(self):
        """
        Test kuebadm.token_generate error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.token_list()

    def test_alpha_certs_renew(self):
        """
        Test kuebadm.alpha_certs_renew without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.alpha_certs_renew() == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "alpha", "certs", "renew"]
            )

    def test_alpha_certs_renew_params(self):
        """
        Test kuebadm.alpha_certs_renew with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.alpha_certs_renew(rootfs="/mnt") == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "alpha", "certs", "renew", "--rootfs", "/mnt"]
            )

    def test_alpha_certs_renew_error(self):
        """
        Test kuebadm.alpha_certs_renew error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.alpha_certs_renew()

    def test_alpha_kubeconfig_user(self):
        """
        Test kuebadm.alpha_kubeconfig_user without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.alpha_kubeconfig_user("user") == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "alpha", "kubeconfig", "user", "--client-name", "user"]
            )

    def test_alpha_kubeconfig_user_params(self):
        """
        Test kuebadm.alpha_kubeconfig_user with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.alpha_kubeconfig_user(
                    "user",
                    apiserver_advertise_address="127.0.0.1",
                    apiserver_bind_port="1234",
                    cert_dir="/pki",
                    org="org",
                    token="token",
                    rootfs="/mnt",
                )
                == "stdout"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "alpha",
                    "kubeconfig",
                    "user",
                    "--client-name",
                    "user",
                    "--apiserver-advertise-address",
                    "127.0.0.1",
                    "--apiserver-bind-port",
                    "1234",
                    "--cert-dir",
                    "/pki",
                    "--org",
                    "org",
                    "--token",
                    "token",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_alpha_kubeconfig_user_error(self):
        """
        Test kuebadm.alpha_kubeconfig_user error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.alpha_kubeconfig_user("user")

    def test_alpha_kubelet_config_download(self):
        """
        Test kuebadm.alpha_kubelet_config_download without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.alpha_kubelet_config_download() == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "alpha", "kubelet", "config", "download"]
            )

    def test_alpha_kubelet_config_download_params(self):
        """
        Test kuebadm.alpha_kubelet_config_download with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.alpha_kubelet_config_download(
                    kubeconfig="/kube.cfg", kubelet_version="version", rootfs="/mnt"
                )
                == "stdout"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "alpha",
                    "kubelet",
                    "config",
                    "download",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--kubelet-version",
                    "version",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_alpha_kubelet_config_download_error(self):
        """
        Test kuebadm.alpha_kubelet_config_download error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.alpha_kubelet_config_download()

    def test_alpha_kubelet_config_enable_dynamic(self):
        """
        Test kuebadm.alpha_kubelet_config_enable_dynamic without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            result = kubeadm.alpha_kubelet_config_enable_dynamic("node-1")
            assert result == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "alpha",
                    "kubelet",
                    "config",
                    "enable-dynamic",
                    "--node-name",
                    "node-1",
                ]
            )

    def test_alpha_kubelet_config_enable_dynamic_params(self):
        """
        Test kuebadm.alpha_kubelet_config_enable_dynamic with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.alpha_kubelet_config_enable_dynamic(
                    "node-1",
                    kubeconfig="/kube.cfg",
                    kubelet_version="version",
                    rootfs="/mnt",
                )
                == "stdout"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "alpha",
                    "kubelet",
                    "config",
                    "enable-dynamic",
                    "--node-name",
                    "node-1",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--kubelet-version",
                    "version",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_alpha_kubelet_config_enable_dynamic_error(self):
        """
        Test kuebadm.alpha_kubelet_config_enable_dynamic error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.alpha_kubelet_config_enable_dynamic("node-1")

    def test_alpha_selfhosting_pivot(self):
        """
        Test kuebadm.alpha_selfhosting_pivot without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.alpha_selfhosting_pivot() == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "alpha", "selfhosting", "pivot", "--force"]
            )

    def test_alpha_selfhosting_pivot_params(self):
        """
        Test kuebadm.alpha_selfhosting_pivot with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.alpha_selfhosting_pivot(
                    cert_dir="/pki",
                    config="/kubeadm.cfg",
                    kubeconfig="/kube.cfg",
                    store_certs_in_secrets=True,
                    rootfs="/mnt",
                )
                == "stdout"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "alpha",
                    "selfhosting",
                    "pivot",
                    "--force",
                    "--store-certs-in-secrets",
                    "--cert-dir",
                    "/pki",
                    "--config",
                    "/kubeadm.cfg",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_alpha_selfhosting_pivot_error(self):
        """
        Test kuebadm.alpha_selfhosting_pivot error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.alpha_selfhosting_pivot()

    def test_config_images_list(self):
        """
        Test kuebadm.config_images_list without parameters
        """
        result = {"retcode": 0, "stdout": "image1\nimage2\n"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.config_images_list() == ["image1", "image2"]
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "config", "images", "list"]
            )

    def test_config_images_list_params(self):
        """
        Test kuebadm.config_images_list with parameters
        """
        result = {"retcode": 0, "stdout": "image1\nimage2\n"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.config_images_list(
                    config="/kubeadm.cfg",
                    feature_gates="k=v",
                    kubernetes_version="version",
                    kubeconfig="/kube.cfg",
                    rootfs="/mnt",
                )
                == ["image1", "image2"]
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "config",
                    "images",
                    "list",
                    "--config",
                    "/kubeadm.cfg",
                    "--feature-gates",
                    "k=v",
                    "--kubernetes-version",
                    "version",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_config_images_list_error(self):
        """
        Test kuebadm.config_images_list error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.config_images_list()

    def test_config_images_pull(self):
        """
        Test kuebadm.config_images_pull without parameters
        """
        result = {
            "retcode": 0,
            "stdout": "[config/images] Pulled image1\n[config/images] Pulled image2\n",
        }
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.config_images_pull() == ["image1", "image2"]
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "config", "images", "pull"]
            )

    def test_config_images_pull_params(self):
        """
        Test kuebadm.config_images_pull with parameters
        """
        result = {
            "retcode": 0,
            "stdout": "[config/images] Pulled image1\n[config/images] Pulled image2\n",
        }
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.config_images_pull(
                    config="/kubeadm.cfg",
                    cri_socket="socket",
                    feature_gates="k=v",
                    kubernetes_version="version",
                    kubeconfig="/kube.cfg",
                    rootfs="/mnt",
                )
                == ["image1", "image2"]
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "config",
                    "images",
                    "pull",
                    "--config",
                    "/kubeadm.cfg",
                    "--cri-socket",
                    "socket",
                    "--feature-gates",
                    "k=v",
                    "--kubernetes-version",
                    "version",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_config_images_pull_error(self):
        """
        Test kuebadm.config_images_pull error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.config_images_pull()

    def test_config_migrate(self):
        """
        Test kuebadm.config_migrate without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.config_migrate("/oldconfig.cfg") == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "config", "migrate", "--old-config", "/oldconfig.cfg"]
            )

    def test_config_migrate_params(self):
        """
        Test kuebadm.config_migrate with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.config_migrate(
                    "/oldconfig.cfg",
                    new_config="/newconfig.cfg",
                    kubeconfig="/kube.cfg",
                    rootfs="/mnt",
                )
                == "stdout"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "config",
                    "migrate",
                    "--old-config",
                    "/oldconfig.cfg",
                    "--new-config",
                    "/newconfig.cfg",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_config_migrate_error(self):
        """
        Test kuebadm.config_migrate error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.config_migrate("/oldconfig.cfg")

    def test_config_print_init_defaults(self):
        """
        Test kuebadm.config_print_init_defaults without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.config_print_init_defaults() == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "config", "print", "init-defaults"]
            )

    def test_config_print_init_defaults_params(self):
        """
        Test kuebadm.config_print_init_defaults with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.config_print_init_defaults(
                    component_configs="component", kubeconfig="/kube.cfg", rootfs="/mnt"
                )
                == "stdout"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "config",
                    "print",
                    "init-defaults",
                    "--component-configs",
                    "component",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_config_print_init_defaults_error(self):
        """
        Test kuebadm.config_print_init_defaults error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.config_print_init_defaults()

    def test_config_print_join_defaults(self):
        """
        Test kuebadm.config_print_join_defaults without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.config_print_join_defaults() == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "config", "print", "join-defaults"]
            )

    def test_config_print_join_defaults_params(self):
        """
        Test kuebadm.config_print_join_defaults with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.config_print_join_defaults(
                    component_configs="component", kubeconfig="/kube.cfg", rootfs="/mnt"
                )
                == "stdout"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "config",
                    "print",
                    "join-defaults",
                    "--component-configs",
                    "component",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_config_print_join_defaults_error(self):
        """
        Test kuebadm.config_print_join_defaults error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.config_print_join_defaults()

    def test_config_upload_from_file(self):
        """
        Test kuebadm.config_upload_from_file without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.config_upload_from_file("/config.cfg") == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "config", "upload", "from-file", "--config", "/config.cfg"]
            )

    def test_config_upload_from_file_params(self):
        """
        Test kuebadm.config_upload_from_file with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.config_upload_from_file(
                    "/config.cfg", kubeconfig="/kube.cfg", rootfs="/mnt"
                )
                == "stdout"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "config",
                    "upload",
                    "from-file",
                    "--config",
                    "/config.cfg",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_config_upload_from_file_error(self):
        """
        Test kuebadm.config_upload_from_file error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.config_upload_from_file("/config.cfg")

    def test_config_upload_from_flags(self):
        """
        Test kuebadm.config_upload_from_flags without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.config_upload_from_flags() == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(
                ["kubeadm", "config", "upload", "from-flags"]
            )

    def test_config_upload_from_flags_params(self):
        """
        Test kuebadm.config_upload_from_flags with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.config_upload_from_flags(
                    apiserver_advertise_address="127.0.0.1",
                    apiserver_bind_port="1234",
                    apiserver_cert_extra_sans="sans",
                    cert_dir="/pki",
                    cri_socket="socket",
                    feature_gates="k=v",
                    kubernetes_version="version",
                    node_name="node-1",
                    pod_network_cidr="10.1.0.0/12",
                    service_cidr="10.2.0.0/12",
                    service_dns_domain="example.org",
                    kubeconfig="/kube.cfg",
                    rootfs="/mnt",
                )
                == "stdout"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "config",
                    "upload",
                    "from-flags",
                    "--apiserver-advertise-address",
                    "127.0.0.1",
                    "--apiserver-bind-port",
                    "1234",
                    "--apiserver-cert-extra-sans",
                    "sans",
                    "--cert-dir",
                    "/pki",
                    "--cri-socket",
                    "socket",
                    "--feature-gates",
                    "k=v",
                    "--kubernetes-version",
                    "version",
                    "--node-name",
                    "node-1",
                    "--pod-network-cidr",
                    "10.1.0.0/12",
                    "--service-cidr",
                    "10.2.0.0/12",
                    "--service-dns-domain",
                    "example.org",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_config_upload_from_flags_error(self):
        """
        Test kuebadm.config_upload_from_flags error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.config_upload_from_flags()

    def test_config_view(self):
        """
        Test kuebadm.config_view without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.config_view() == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(["kubeadm", "config", "view"])

    def test_config_view_params(self):
        """
        Test kuebadm.config_view with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.config_view(kubeconfig="/kube.cfg", rootfs="/mnt") == "stdout"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "config",
                    "view",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_config_view_error(self):
        """
        Test kuebadm.config_view error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.config_view()

    def test_init(self):
        """
        Test kuebadm.init without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.init() == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(["kubeadm", "init"])

    def test_init_params(self):
        """
        Test kuebadm.init with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.init(
                    apiserver_advertise_address="127.0.0.1",
                    apiserver_bind_port="1234",
                    apiserver_cert_extra_sans="sans",
                    cert_dir="/pki",
                    certificate_key="secret",
                    config="/config.cfg",
                    cri_socket="socket",
                    experimental_upload_certs=True,
                    feature_gates="k=v",
                    ignore_preflight_errors="all",
                    image_repository="example.org",
                    kubernetes_version="version",
                    node_name="node-1",
                    pod_network_cidr="10.1.0.0/12",
                    service_cidr="10.2.0.0/12",
                    service_dns_domain="example.org",
                    skip_certificate_key_print=True,
                    skip_phases="all",
                    skip_token_print=True,
                    token="token",
                    token_ttl="1h1m1s",
                    rootfs="/mnt",
                )
                == "stdout"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "init",
                    "--experimental-upload-certs",
                    "--skip-certificate-key-print",
                    "--skip-token-print",
                    "--apiserver-advertise-address",
                    "127.0.0.1",
                    "--apiserver-bind-port",
                    "1234",
                    "--apiserver-cert-extra-sans",
                    "sans",
                    "--cert-dir",
                    "/pki",
                    "--certificate-key",
                    "secret",
                    "--config",
                    "/config.cfg",
                    "--cri-socket",
                    "socket",
                    "--feature-gates",
                    "k=v",
                    "--ignore-preflight-errors",
                    "all",
                    "--image-repository",
                    "example.org",
                    "--kubernetes-version",
                    "version",
                    "--node-name",
                    "node-1",
                    "--pod-network-cidr",
                    "10.1.0.0/12",
                    "--service-cidr",
                    "10.2.0.0/12",
                    "--service-dns-domain",
                    "example.org",
                    "--skip-phases",
                    "all",
                    "--token",
                    "token",
                    "--token-ttl",
                    "1h1m1s",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_init_error(self):
        """
        Test kuebadm.init error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.init()

    def test_join(self):
        """
        Test kuebadm.join without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.join() == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(["kubeadm", "join"])

    def test_join_params(self):
        """
        Test kuebadm.join with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.join(
                    api_server_endpoint="10.160.65.165:6443",
                    apiserver_advertise_address="127.0.0.1",
                    apiserver_bind_port="1234",
                    certificate_key="secret",
                    config="/config.cfg",
                    cri_socket="socket",
                    discovery_file="/discovery.cfg",
                    discovery_token="token",
                    discovery_token_ca_cert_hash="type:value",
                    discovery_token_unsafe_skip_ca_verification=True,
                    experimental_control_plane=True,
                    ignore_preflight_errors="all",
                    node_name="node-1",
                    skip_phases="all",
                    tls_bootstrap_token="token",
                    token="token",
                    rootfs="/mnt",
                )
                == "stdout"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "join",
                    "10.160.65.165:6443",
                    "--discovery-token-unsafe-skip-ca-verification",
                    "--experimental-control-plane",
                    "--apiserver-advertise-address",
                    "127.0.0.1",
                    "--apiserver-bind-port",
                    "1234",
                    "--certificate-key",
                    "secret",
                    "--config",
                    "/config.cfg",
                    "--cri-socket",
                    "socket",
                    "--discovery-file",
                    "/discovery.cfg",
                    "--discovery-token",
                    "token",
                    "--discovery-token-ca-cert-hash",
                    "type:value",
                    "--ignore-preflight-errors",
                    "all",
                    "--node-name",
                    "node-1",
                    "--skip-phases",
                    "all",
                    "--tls-bootstrap-token",
                    "token",
                    "--token",
                    "token",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_join_error(self):
        """
        Test kuebadm.join error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.join()

    def test_reset(self):
        """
        Test kuebadm.reset without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert kubeadm.reset() == "stdout"
            salt_mock["cmd.run_all"].assert_called_with(["kubeadm", "reset", "--force"])

    def test_reset_params(self):
        """
        Test kuebadm.reset with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            assert (
                kubeadm.reset(
                    cert_dir="/pki",
                    cri_socket="socket",
                    ignore_preflight_errors="all",
                    kubeconfig="/kube.cfg",
                    rootfs="/mnt",
                )
                == "stdout"
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "kubeadm",
                    "reset",
                    "--force",
                    "--cert-dir",
                    "/pki",
                    "--cri-socket",
                    "socket",
                    "--ignore-preflight-errors",
                    "all",
                    "--kubeconfig",
                    "/kube.cfg",
                    "--rootfs",
                    "/mnt",
                ]
            )

    def test_reset_error(self):
        """
        Test kuebadm.reset error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(kubeadm.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert kubeadm.reset()



"""
Grains from cloud metadata servers at 169.254.169.254

.. versionadded:: 2017.7.0

:depends: requests

To enable these grains that pull from the http://169.254.169.254/latest
metadata server set `metadata_server_grains: True` in the minion config.

.. code-block:: yaml

    metadata_server_grains: True

"""

import os
import socket

import salt.utils.data
import salt.utils.http as http
import salt.utils.json
import salt.utils.stringutils

# metadata server information
IP = "169.254.169.254"
HOST = "http://{}/".format(IP)


def __virtual__():
    if __opts__.get("metadata_server_grains", False) is False:
        return False
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(0.1)
    result = sock.connect_ex((IP, 80))
    if result != 0:
        return False
    if http.query(os.path.join(HOST, "latest/"), status=True).get("status") != 200:
        return False
    return True


def _search(prefix="latest/"):
    """
    Recursively look up all grains in the metadata server
    """
    ret = {}
    linedata = http.query(os.path.join(HOST, prefix), headers=True)
    if "body" not in linedata:
        return ret
    body = salt.utils.stringutils.to_unicode(linedata["body"])
    if (
        linedata["headers"].get("Content-Type", "text/plain")
        == "application/octet-stream"
    ):
        return body
    for line in body.split("\n"):
        if line.endswith("/"):
            ret[line[:-1]] = _search(prefix=os.path.join(prefix, line))
        elif prefix == "latest/":
            # (gtmanfred) The first level should have a forward slash since
            # they have stuff underneath. This will not be doubled up though,
            # because lines ending with a slash are checked first.
            ret[line] = _search(prefix=os.path.join(prefix, line + "/"))
        elif line.endswith(("dynamic", "meta-data")):
            ret[line] = _search(prefix=os.path.join(prefix, line))
        elif "=" in line:
            key, value = line.split("=")
            ret[value] = _search(prefix=os.path.join(prefix, key))
        else:
            retdata = http.query(os.path.join(HOST, prefix, line)).get("body", None)
            # (gtmanfred) This try except block is slightly faster than
            # checking if the string starts with a curly brace
            if isinstance(retdata, bytes):
                try:
                    ret[line] = salt.utils.json.loads(
                        salt.utils.stringutils.to_unicode(retdata)
                    )
                except ValueError:
                    ret[line] = salt.utils.stringutils.to_unicode(retdata)
            else:
                ret[line] = retdata
    return salt.utils.data.decode(ret)


def metadata():
    return _search()



"""
The service module for NetBSD

.. important::
    If you feel that Salt should be using this module to manage services on a
    minion, and it is using a different module (or gives an error similar to
    *'service.start' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import fnmatch
import glob
import os
import re

__func_alias__ = {"reload_": "reload"}

# Define the module's virtual name
__virtualname__ = "service"


def __virtual__():
    """
    Only work on NetBSD
    """
    if __grains__["os"] == "NetBSD" and os.path.exists("/etc/rc.subr"):
        return __virtualname__
    return (
        False,
        "The netbsdservice execution module failed to load: only available on NetBSD.",
    )


def start(name):
    """
    Start the specified service

    CLI Example:

    .. code-block:: bash

        salt '*' service.start <service name>
    """
    cmd = "/etc/rc.d/{} onestart".format(name)
    return not __salt__["cmd.retcode"](cmd)


def stop(name):
    """
    Stop the specified service

    CLI Example:

    .. code-block:: bash

        salt '*' service.stop <service name>
    """
    cmd = "/etc/rc.d/{} onestop".format(name)
    return not __salt__["cmd.retcode"](cmd)


def restart(name):
    """
    Restart the named service

    CLI Example:

    .. code-block:: bash

        salt '*' service.restart <service name>
    """
    cmd = "/etc/rc.d/{} onerestart".format(name)
    return not __salt__["cmd.retcode"](cmd)


def reload_(name):
    """
    Reload the named service

    CLI Example:

    .. code-block:: bash

        salt '*' service.reload <service name>
    """
    cmd = "/etc/rc.d/{} onereload".format(name)
    return not __salt__["cmd.retcode"](cmd)


def force_reload(name):
    """
    Force-reload the named service

    CLI Example:

    .. code-block:: bash

        salt '*' service.force_reload <service name>
    """
    cmd = "/etc/rc.d/{} forcereload".format(name)
    return not __salt__["cmd.retcode"](cmd)


def status(name, sig=None):
    """
    Return the status for a service.
    If the name contains globbing, a dict mapping service name to True/False
    values is returned.

    .. versionchanged:: 2018.3.0
        The service name can now be a glob (e.g. ``salt*``)

    Args:
        name (str): The name of the service to check
        sig (str): Signature to use to find the service via ps

    Returns:
        bool: True if running, False otherwise
        dict: Maps service name to True if running, False otherwise

    CLI Example:

    .. code-block:: bash

        salt '*' service.status <service name> [service signature]
    """
    if sig:
        return bool(__salt__["status.pid"](sig))

    contains_globbing = bool(re.search(r"\*|\?|\[.+\]", name))
    if contains_globbing:
        services = fnmatch.filter(get_all(), name)
    else:
        services = [name]
    results = {}
    for service in services:
        cmd = "/etc/rc.d/{} onestatus".format(service)
        results[service] = not __salt__["cmd.retcode"](cmd, ignore_retcode=True)
    if contains_globbing:
        return results
    return results[name]


def _get_svc(rcd, service_status):
    """
    Returns a unique service status
    """
    ena = None
    lines = __salt__["cmd.run"]("{} rcvar".format(rcd)).splitlines()
    for rcvar in lines:
        if rcvar.startswith("$") and "={}".format(service_status) in rcvar:
            ena = "yes"
        elif rcvar.startswith("#"):
            svc = rcvar.split(" ", 1)[1]
        else:
            continue

    if ena and svc:
        return svc
    return None


def _get_svc_list(service_status):
    """
    Returns all service statuses
    """
    prefix = "/etc/rc.d/"
    ret = set()
    lines = glob.glob("{}*".format(prefix))
    for line in lines:
        svc = _get_svc(line, service_status)
        if svc is not None:
            ret.add(svc)

    return sorted(ret)


def get_enabled():
    """
    Return a list of service that are enabled on boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_enabled
    """
    return _get_svc_list("YES")


def get_disabled():
    """
    Return a set of services that are installed but disabled

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_disabled
    """
    return _get_svc_list("NO")


def available(name):
    """
    Returns ``True`` if the specified service is available, otherwise returns
    ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' service.available sshd
    """
    return name in get_all()


def missing(name):
    """
    The inverse of service.available.
    Returns ``True`` if the specified service is not available, otherwise returns
    ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' service.missing sshd
    """
    return name not in get_all()


def get_all():
    """
    Return all available boot services

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_all
    """
    return _get_svc_list("")


def _rcconf_status(name, service_status):
    """
    Modifies /etc/rc.conf so a service is started or not at boot time and
    can be started via /etc/rc.d/<service>
    """
    rcconf = "/etc/rc.conf"
    rxname = "^{}=.*".format(name)
    newstatus = "{}={}".format(name, service_status)
    ret = __salt__["cmd.retcode"]("grep '{}' {}".format(rxname, rcconf))
    if ret == 0:  # service found in rc.conf, modify its status
        __salt__["file.replace"](rcconf, rxname, newstatus)
    else:
        ret = __salt__["file.append"](rcconf, newstatus)

    return ret


def enable(name, **kwargs):
    """
    Enable the named service to start at boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.enable <service name>
    """
    return _rcconf_status(name, "YES")


def disable(name, **kwargs):
    """
    Disable the named service to start at boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.disable <service name>
    """
    return _rcconf_status(name, "NO")


def enabled(name, **kwargs):
    """
    Return True if the named service is enabled, false otherwise

    CLI Example:

    .. code-block:: bash

        salt '*' service.enabled <service name>
    """
    return _get_svc("/etc/rc.d/{}".format(name), "YES")


def disabled(name):
    """
    Return True if the named service is enabled, false otherwise

    CLI Example:

    .. code-block:: bash

        salt '*' service.disabled <service name>
    """
    return _get_svc("/etc/rc.d/{}".format(name), "NO")


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



"""
    tests.unit.returners.cassandra_cql_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""


import ssl

import salt.modules.cassandra_cql as cassandra_cql
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

try:
    import cassandra  # pylint: disable=unused-import,wrong-import-position

    HAS_CASSANDRA = True
except ImportError:
    HAS_CASSANDRA = False


@skipIf(
    not HAS_CASSANDRA,
    "Please install the cassandra datastax driver to run cassandra_cql module unit"
    " tests.",
)
class CassandraCQLReturnerTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cassandra CQL module
    """

    def setup_loader_modules(self):
        return {cassandra_cql: {}}

    def test_returns_opts_if_specified(self):
        """
        If ssl options are present then check that they are parsed and returned
        """
        options = MagicMock(
            return_value={
                "cluster": ["192.168.50.10", "192.168.50.11", "192.168.50.12"],
                "port": 9000,
                "ssl_options": {
                    "ca_certs": "/etc/ssl/certs/ca-bundle.trust.crt",
                    "ssl_version": "PROTOCOL_TLSv1",
                },
                "username": "cas_admin",
            }
        )

        with patch.dict(cassandra_cql.__salt__, {"config.option": options}):

            self.assertEqual(
                cassandra_cql._get_ssl_opts(),
                {  # pylint: disable=protected-access
                    "ca_certs": "/etc/ssl/certs/ca-bundle.trust.crt",
                    "ssl_version": ssl.PROTOCOL_TLSv1,
                },
            )  # pylint: disable=no-member

    def test_invalid_protocol_version(self):
        """
        Check that the protocol version is imported only if it isvalid
        """
        options = MagicMock(
            return_value={
                "cluster": ["192.168.50.10", "192.168.50.11", "192.168.50.12"],
                "port": 9000,
                "ssl_options": {
                    "ca_certs": "/etc/ssl/certs/ca-bundle.trust.crt",
                    "ssl_version": "Invalid",
                },
                "username": "cas_admin",
            }
        )

        with patch.dict(cassandra_cql.__salt__, {"config.option": options}):
            with self.assertRaises(CommandExecutionError):
                cassandra_cql._get_ssl_opts()  # pylint: disable=protected-access

    def test_unspecified_opts(self):
        """
        Check that it returns None when ssl opts aren't specified
        """
        with patch.dict(
            cassandra_cql.__salt__, {"config.option": MagicMock(return_value={})}
        ):
            self.assertEqual(
                cassandra_cql._get_ssl_opts(), None  # pylint: disable=protected-access
            )

    def test_valid_asynchronous_args(self):
        mock_execute = MagicMock(return_value={})
        mock_execute_async = MagicMock(return_value={})
        mock_context = {
            "cassandra_cql_returner_cluster": MagicMock(return_value={}),
            "cassandra_cql_returner_session": MagicMock(
                execute=mock_execute,
                execute_async=mock_execute_async,
                prepare=lambda _: MagicMock(
                    bind=lambda _: None
                ),  # mock prepared_statement
                row_factory=None,
            ),
            "cassandra_cql_prepared": {},
        }

        with patch.dict(cassandra_cql.__context__, mock_context):
            cassandra_cql.cql_query_with_prepare(
                "SELECT now() from system.local;", "select_now", [], asynchronous=True
            )
            mock_execute_async.assert_called_once()

    def test_valid_async_args(self):
        mock_execute = MagicMock(return_value={})
        mock_execute_async = MagicMock(return_value={})
        mock_context = {
            "cassandra_cql_returner_cluster": MagicMock(return_value={}),
            "cassandra_cql_returner_session": MagicMock(
                execute=mock_execute,
                execute_async=mock_execute_async,
                prepare=lambda _: MagicMock(bind=lambda _: None),
                # mock prepared_statement
                row_factory=None,
            ),
            "cassandra_cql_prepared": {},
        }

        kwargs = {"async": True}  # to avoid syntax error in python 3.7
        with patch.dict(cassandra_cql.__context__, mock_context):
            cassandra_cql.cql_query_with_prepare(
                "SELECT now() from system.local;", "select_now", [], **kwargs
            )
            mock_execute_async.assert_called_once()



#
# Author: Bo Maryniuk <bo@suse.de>
#
# Copyright 2017 SUSE LLC
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Salt Service Discovery Protocol.
JSON-based service discovery protocol, used by minions to find running Master.
"""

import copy
import datetime
import logging
import random
import socket
import time

import salt.utils.json
import salt.utils.stringutils

try:
    from salt.utils.odict import OrderedDict
except ImportError:
    from collections import OrderedDict

_json = salt.utils.json.import_json()
if not hasattr(_json, "dumps"):
    _json = None

try:
    import asyncio

    asyncio.ported = False
except ImportError:
    try:
        # Python 2 doesn't have asyncio
        import trollius as asyncio

        asyncio.ported = True
    except ImportError:
        asyncio = None


class TimeOutException(Exception):
    pass


class TimeStampException(Exception):
    pass


class SSDPBase:
    """
    Salt Service Discovery Protocol.
    """

    log = logging.getLogger(__name__)

    # Fields
    SIGNATURE = "signature"
    ANSWER = "answer"
    PORT = "port"
    LISTEN_IP = "listen_ip"
    TIMEOUT = "timeout"

    # Default values
    DEFAULTS = {
        SIGNATURE: "__salt_master_service",
        PORT: 4520,
        LISTEN_IP: "0.0.0.0",
        TIMEOUT: 3,
        ANSWER: {},
    }

    @staticmethod
    def _is_available():
        """
        Return True if the USSDP dependencies are satisfied.
        :return:
        """
        return bool(asyncio and _json)

    @staticmethod
    def get_self_ip():
        """
        Find out localhost outside IP.

        :return:
        """
        sck = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            sck.connect(("1.255.255.255", 1))  # Does not needs to be reachable
            ip_addr = sck.getsockname()[0]
        except Exception:  # pylint: disable=broad-except
            ip_addr = socket.gethostbyname(socket.gethostname())
        finally:
            sck.close()
        return ip_addr


class SSDPFactory(SSDPBase):
    """
    Socket protocol factory.
    """

    def __init__(self, **config):
        """
        Initialize

        :param config:
        """
        for attr in (self.SIGNATURE, self.ANSWER):
            setattr(self, attr, config.get(attr, self.DEFAULTS[attr]))
        self.disable_hidden = False
        self.transport = None
        self.my_ip = socket.gethostbyname(socket.gethostname())

    def __call__(self, *args, **kwargs):
        """
        Return instance on Factory call.
        :param args:
        :param kwargs:
        :return:
        """
        return self

    def connection_made(self, transport):
        """
        On connection.

        :param transport:
        :return:
        """
        self.transport = transport

    def _sendto(self, data, addr=None, attempts=10):
        """
        On multi-master environments, running on the same machine,
        transport sending to the destination can be allowed only at once.
        Since every machine will immediately respond, high chance to
        get sending fired at the same time, which will result to a PermissionError
        at socket level. We are attempting to send it in a different time.

        :param data:
        :param addr:
        :return:
        """
        tries = 0
        slp_time = lambda: 0.5 / random.randint(10, 30)
        slp = slp_time()
        while tries < attempts:
            try:
                self.transport.sendto(data, addr=addr)
                self.log.debug("Sent successfully")
                return
            except AttributeError as ex:
                self.log.debug("Permission error: %s", ex)
                time.sleep(slp)
                tries += 1
                slp += slp_time()

    def datagram_received(self, data, addr):
        """
        On datagram receive.

        :param data:
        :param addr:
        :return:
        """
        message = salt.utils.stringutils.to_unicode(data)
        if message.startswith(self.signature):
            try:
                timestamp = float(message[len(self.signature) :])
            except (TypeError, ValueError):
                self.log.debug(
                    "Received invalid timestamp in package from %s:%s", *addr
                )
                if self.disable_hidden:
                    self._sendto(
                        "{}:E:{}".format(self.signature, "Invalid timestamp"), addr
                    )
                return

            if datetime.datetime.fromtimestamp(timestamp) < (
                datetime.datetime.now() - datetime.timedelta(seconds=20)
            ):
                if self.disable_hidden:
                    self._sendto(
                        "{}:E:{}".format(self.signature, "Timestamp is too old"), addr
                    )
                self.log.debug("Received outdated package from %s:%s", *addr)
                return

            self.log.debug('Received "%s" from %s:%s', message, *addr)
            self._sendto(
                salt.utils.stringutils.to_bytes(
                    "{}:@:{}".format(
                        self.signature,
                        salt.utils.json.dumps(self.answer, _json_module=_json),
                    )
                ),
                addr,
            )
        else:
            if self.disable_hidden:
                self._sendto(
                    salt.utils.stringutils.to_bytes(
                        "{}:E:{}".format(self.signature, "Invalid packet signature"),
                        addr,
                    )
                )
            self.log.debug("Received bad signature from %s:%s", *addr)


class SSDPDiscoveryServer(SSDPBase):
    """
    Discovery service publisher.

    """

    @staticmethod
    def is_available():
        """
        Return availability of the Server.
        :return:
        """
        return SSDPBase._is_available()

    def __init__(self, **config):
        """
        Initialize.

        :param config:
        """
        self._config = copy.deepcopy(config)
        if self.ANSWER not in self._config:
            self._config[self.ANSWER] = {}
        self._config[self.ANSWER].update({"master": self.get_self_ip()})

    @staticmethod
    def create_datagram_endpoint(
        loop,
        protocol_factory,
        local_addr=None,
        remote_addr=None,
        family=0,
        proto=0,
        flags=0,
    ):
        """
        Create datagram connection.

        Based on code from Python 3.5 version, this method is used
        only in Python 2.7+ versions, since Trollius library did not
        ported UDP packets broadcast.
        """
        if not (local_addr or remote_addr):
            if not family:
                raise ValueError("unexpected address family")
            addr_pairs_info = (((family, proto), (None, None)),)
        else:
            addr_infos = OrderedDict()
            for idx, addr in ((0, local_addr), (1, remote_addr)):
                if addr is not None:
                    assert (
                        isinstance(addr, tuple) and len(addr) == 2
                    ), "2-tuple is expected"
                    infos = yield asyncio.coroutines.From(
                        loop.getaddrinfo(
                            *addr,
                            family=family,
                            type=socket.SOCK_DGRAM,
                            proto=proto,
                            flags=flags
                        )
                    )
                    if not infos:
                        raise OSError("getaddrinfo() returned empty list")
                    for fam, _, pro, _, address in infos:
                        key = (fam, pro)
                        if key not in addr_infos:
                            addr_infos[key] = [None, None]
                        addr_infos[key][idx] = address
            addr_pairs_info = [
                (key, addr_pair)
                for key, addr_pair in addr_infos.items()
                if not (
                    (local_addr and addr_pair[0] is None)
                    or (remote_addr and addr_pair[1] is None)
                )
            ]
            if not addr_pairs_info:
                raise ValueError("can not get address information")
        exceptions = []
        for ((family, proto), (local_address, remote_address)) in addr_pairs_info:
            sock = r_addr = None
            try:
                sock = socket.socket(family=family, type=socket.SOCK_DGRAM, proto=proto)
                for opt in [socket.SO_REUSEADDR, socket.SO_BROADCAST]:
                    sock.setsockopt(socket.SOL_SOCKET, opt, 1)
                sock.setblocking(False)
                if local_addr:
                    sock.bind(local_address)
                if remote_addr:
                    yield asyncio.coroutines.From(
                        loop.sock_connect(sock, remote_address)
                    )
                    r_addr = remote_address
            except OSError as exc:
                if sock is not None:
                    sock.close()
                exceptions.append(exc)
            except Exception:  # pylint: disable=broad-except
                if sock is not None:
                    sock.close()
                raise
            else:
                break
        else:
            raise exceptions[0]

        protocol = protocol_factory()
        waiter = asyncio.futures.Future(loop=loop)
        transport = loop._make_datagram_transport(sock, protocol, r_addr, waiter)
        try:
            yield asyncio.coroutines.From(waiter)
        except Exception:  # pylint: disable=broad-except
            transport.close()
            raise
        raise asyncio.coroutines.Return(transport, protocol)

    def run(self):
        """
        Run server.
        :return:
        """
        listen_ip = self._config.get(self.LISTEN_IP, self.DEFAULTS[self.LISTEN_IP])
        port = self._config.get(self.PORT, self.DEFAULTS[self.PORT])
        self.log.info(
            "Starting service discovery listener on udp://%s:%s", listen_ip, port
        )
        loop = asyncio.get_event_loop()
        protocol = SSDPFactory(answer=self._config[self.ANSWER])
        if asyncio.ported:
            transport, protocol = loop.run_until_complete(
                SSDPDiscoveryServer.create_datagram_endpoint(
                    loop, protocol, local_addr=(listen_ip, port)
                )
            )
        else:
            transport, protocol = loop.run_until_complete(
                loop.create_datagram_endpoint(
                    protocol, local_addr=(listen_ip, port), allow_broadcast=True
                )
            )
        try:
            loop.run_forever()
        finally:
            self.log.info("Stopping service discovery listener.")
            transport.close()
            loop.close()


class SSDPDiscoveryClient(SSDPBase):
    """
    Class to discover Salt Master via UDP broadcast.
    """

    @staticmethod
    def is_available():
        """
        Return availability of the Client
        :return:
        """
        return SSDPBase._is_available()

    def __init__(self, **config):
        """
        Initialize
        """
        self._config = config
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self._socket.settimeout(
            self._config.get(self.TIMEOUT, self.DEFAULTS[self.TIMEOUT])
        )

        for attr in [self.SIGNATURE, self.TIMEOUT, self.PORT]:
            setattr(self, attr, self._config.get(attr, self.DEFAULTS[attr]))

    def _query(self):
        """
        Query the broadcast for defined services.
        :return:
        """
        query = salt.utils.stringutils.to_bytes(
            "{}{}".format(self.signature, time.time())
        )
        self._socket.sendto(query, ("<broadcast>", self.port))

        return query

    def _collect_masters_map(self, response):
        """
        Collect masters map from the network.
        :return:
        """
        while True:
            try:
                data, addr = self._socket.recvfrom(0x400)
                if data:
                    if addr not in response:
                        response[addr] = []
                    response[addr].append(data)
                else:
                    break
            except Exception as err:  # pylint: disable=broad-except
                self.log.error("Discovery master collection failure: %s", err)
                break

    def discover(self):
        """
        Gather the information of currently declared servers.

        :return:
        """
        response = {}
        masters = {}
        self.log.info("Looking for a server discovery")
        self._query()
        self._collect_masters_map(response)
        if not response:
            msg = "No master has been discovered."
            self.log.info(msg)
        else:
            for addr, descriptions in response.items():
                for (
                    data
                ) in descriptions:  # Several masters can run at the same machine.
                    msg = salt.utils.stringutils.to_unicode(data)
                    if msg.startswith(self.signature):
                        msg = msg.split(self.signature)[-1]
                        self.log.debug(
                            "Service announcement at '%s:%s'. Response: '%s'",
                            addr[0],
                            addr[1],
                            msg,
                        )
                        if ":E:" in msg:
                            err = msg.split(":E:")[-1]
                            self.log.error(
                                "Error response from the service publisher at %s: %s",
                                addr,
                                err,
                            )
                            if "timestamp" in err:
                                self.log.error(
                                    "Publisher sent shifted timestamp from %s", addr
                                )
                        else:
                            if addr not in masters:
                                masters[addr] = []
                            masters[addr].append(
                                salt.utils.json.loads(
                                    msg.split(":@:")[-1], _json_module=_json
                                )
                            )
        return masters



"""
Tests for building and installing salt
"""
import json
import logging
import pathlib
import re
import sys

import pytest
import salt.utils.path
import salt.utils.platform
import salt.version
from salt.modules.virtualenv_mod import KNOWN_BINARY_NAMES

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_if_binaries_missing(*KNOWN_BINARY_NAMES, check_all=False),
]


def use_static_requirements_ids(value):
    return "USE_STATIC_REQUIREMENTS={}".format("1" if value else "0")


@pytest.fixture(params=[True, False], ids=use_static_requirements_ids)
def use_static_requirements(request):
    return request.param


@pytest.fixture
def virtualenv(virtualenv, use_static_requirements):
    virtualenv.environ["USE_STATIC_REQUIREMENTS"] = (
        "1" if use_static_requirements else "0"
    )
    return virtualenv


def test_wheel(virtualenv, cache_dir, use_static_requirements, src_dir):
    """
    test building and installing a bdist_wheel package
    """
    # Let's create the testing virtualenv
    with virtualenv as venv:
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)
        venv.run(
            venv.venv_python,
            "setup.py",
            "bdist_wheel",
            "--dist-dir",
            str(cache_dir),
            cwd=src_dir,
        )
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        salt_generated_package = list(cache_dir.glob("*.whl"))
        if not salt_generated_package:
            pytest.fail("Could not find the generated wheel file")
        salt_generated_package = salt_generated_package[0]

        # Assert generate wheel version matches what salt reports as its version
        whl_ver = [
            x
            for x in salt_generated_package.name.split("-")
            if re.search(r"^\d.\d*", x)
        ][0]
        whl_ver_cmp = whl_ver.replace("_", "-")
        assert whl_ver_cmp == salt.version.__version__, "{} != {}".format(
            whl_ver_cmp, salt.version.__version__
        )

        # Because bdist_wheel supports pep517, we don't have to pre-install Salt's
        # dependencies before installing the wheel package
        if not use_static_requirements and salt.utils.platform.is_windows():
            # However, on windows, the latest pycurl release, 7.43.0.6 at the time of writing,
            # does not have wheel files uploaded, so, we force pycurl==7.43.0.5 to be
            # pre-installed before installing salt
            venv.install("pycurl==7.43.0.5")
        venv.install(str(salt_generated_package))

        # Let's ensure the version is correct
        cmd = venv.run(venv.venv_python, "-m", "pip", "list", "--format", "json")
        for details in json.loads(cmd.stdout):
            if details["name"] != "salt":
                continue
            installed_version = details["version"]
            break
        else:
            pytest.fail("Salt was not found installed")

        # Let's compare the installed version with the version salt reports
        assert installed_version == salt.version.__version__, "{} != {}".format(
            installed_version, salt.version.__version__
        )

        # Let's also ensure we have a salt/_version.py from the installed salt wheel
        subdir = [
            "lib",
            "python{}.{}".format(*sys.version_info),
            "site-packages",
            "salt",
        ]
        if salt.utils.platform.is_windows():
            subdir.pop(1)

        installed_salt_path = pathlib.Path(venv.venv_dir)
        installed_salt_path = installed_salt_path.joinpath(*subdir)
        assert installed_salt_path.is_dir()
        salt_generated_version_file_path = installed_salt_path / "_version.py"
        assert salt_generated_version_file_path.is_file()


def test_egg(virtualenv, cache_dir, use_static_requirements, src_dir):
    """
    test building and installing a bdist_egg package
    """
    # TODO: We should actually disallow generating an egg file
    # Let's create the testing virtualenv
    with virtualenv as venv:
        ret = venv.run(
            venv.venv_python,
            "-c",
            "import setuptools; print(setuptools.__version__)",
        )
        setuptools_version = ret.stdout.strip()
        ret = venv.run(venv.venv_python, "-m", "easy_install", "--version", check=False)
        if ret.exitcode != 0:
            pytest.skip(
                "Setuptools version, {}, does not include the easy_install module".format(
                    setuptools_version
                )
            )
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        # Setuptools installs pre-release packages if we don't pin to an exact version
        # Let's download and install requirements before, running salt's install test
        venv.run(
            venv.venv_python,
            "-m",
            "pip",
            "download",
            "--dest",
            str(cache_dir),
            src_dir,
        )
        packages = []
        for fname in cache_dir.iterdir():
            if (
                fname.name.startswith("pycurl")
                and salt.utils.platform.is_windows()
                and not use_static_requirements
            ):
                # On windows, the latest pycurl release, 7.43.0.6 at the time of writing,
                # does not have wheel files uploaded, so, delete the downloaded source
                # tarball and will later force pycurl==7.43.0.5 to be pre-installed before
                # installing salt
                fname.unlink()
                continue
            packages.append(fname)
        venv.install(*[str(pkg) for pkg in packages])
        for package in packages:
            package.unlink()

        # Looks like, at least on windows, setuptools also get's downloaded as a salt dependency.
        # Let's check and see if this newly installed version also has easy_install
        ret = venv.run(
            venv.venv_python,
            "-c",
            "import setuptools; print(setuptools.__version__)",
        )
        setuptools_version = ret.stdout.strip()
        ret = venv.run(venv.venv_python, "-m", "easy_install", "--version", check=False)
        if ret.exitcode != 0:
            pytest.skip(
                "Setuptools version, {}, does not include the easy_install module".format(
                    setuptools_version
                )
            )

        if salt.utils.platform.is_windows() and not use_static_requirements:
            # Like mentioned above, install pycurl==7.43.0.5
            # However, on windows, the latest pycurl release, 7.43.0.6 at the time of writing,
            # does not have wheel files uploaded, so, we force pycurl==7.43.0.5 to be
            # pre-installed before installing salt
            venv.install("pycurl==7.43.0.5")

        venv.run(
            venv.venv_python,
            "setup.py",
            "bdist_egg",
            "--dist-dir",
            str(cache_dir),
            cwd=src_dir,
        )
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        salt_generated_package = list(cache_dir.glob("*.egg"))
        if not salt_generated_package:
            pytest.fail("Could not find the generated egg file")
        salt_generated_package = salt_generated_package[0]

        # Assert generate wheel version matches what salt reports as its version
        egg_ver = [
            x
            for x in salt_generated_package.name.split("-")
            if re.search(r"^\d.\d*", x)
        ][0]
        egg_ver_cmp = egg_ver.replace("_", "-")
        assert egg_ver_cmp == salt.version.__version__, "{} != {}".format(
            egg_ver_cmp, salt.version.__version__
        )

        # We cannot pip install an egg file, let's go old school
        venv.run(venv.venv_python, "-m", "easy_install", str(salt_generated_package))

        # Let's ensure the version is correct
        cmd = venv.run(venv.venv_python, "-m", "pip", "list", "--format", "json")
        for details in json.loads(cmd.stdout):
            if details["name"] != "salt":
                continue
            installed_version = details["version"]
            break
        else:
            pytest.fail("Salt was not found installed")

        # Let's compare the installed version with the version salt reports
        assert installed_version == salt.version.__version__, "{} != {}".format(
            installed_version, salt.version.__version__
        )

        # Let's also ensure we have a salt/_version.py from the installed salt egg
        subdir = [
            "lib",
            "python{}.{}".format(*sys.version_info),
            "site-packages",
        ]
        if salt.utils.platform.is_windows():
            subdir.pop(1)
        site_packages_dir = pathlib.Path(venv.venv_dir)
        site_packages_dir = site_packages_dir.joinpath(*subdir)
        assert site_packages_dir.is_dir()
        installed_salt_path = list(site_packages_dir.glob("salt*.egg"))
        if not installed_salt_path:
            pytest.fail("Failed to find the installed salt path")
        log.debug("Installed salt path glob matches: %s", installed_salt_path)
        installed_salt_path = installed_salt_path[0] / "salt"
        assert installed_salt_path.is_dir()
        salt_generated_version_file_path = installed_salt_path / "_version.py"
        assert salt_generated_version_file_path.is_file(), "{} is not a file".format(
            salt_generated_version_file_path
        )


# On python 3.5 Windows sdist fails with encoding errors. This is resolved
# in later versions.
@pytest.mark.skipif(
    salt.utils.platform.is_windows()
    and sys.version_info > (3,)
    and sys.version_info < (3, 6),
    reason="Skip on python 3.5",
)
def test_sdist(virtualenv, cache_dir, use_static_requirements, src_dir):
    """
    test building and installing a sdist package
    """
    # Let's create the testing virtualenv
    with virtualenv as venv:
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        # Setuptools installs pre-release packages if we don't pin to an exact version
        # Let's download and install requirements before, running salt's install test
        venv.run(
            venv.venv_python,
            "-m",
            "pip",
            "download",
            "--dest",
            str(cache_dir),
            src_dir,
        )
        packages = []
        for fname in cache_dir.iterdir():
            if (
                fname.name.startswith("pycurl")
                and salt.utils.platform.is_windows()
                and not use_static_requirements
            ):
                # On windows, the latest pycurl release, 7.43.0.6 at the time of writing,
                # does not have wheel files uploaded, so, delete the downloaded source
                # tarball and will later force pycurl==7.43.0.5 to be pre-installed before
                # installing salt
                fname.unlink()
                continue
            packages.append(fname)
        venv.install(*[str(pkg) for pkg in packages])
        for package in packages:
            package.unlink()

        if salt.utils.platform.is_windows() and not use_static_requirements:
            # Like mentioned above, install pycurl==7.43.0.5
            # However, on windows, the latest pycurl release, 7.43.0.6 at the time of writing,
            # does not have wheel files uploaded, so, we force pycurl==7.43.0.5 to be
            # pre-installed before installing salt
            venv.install("pycurl==7.43.0.5")

        venv.run(
            venv.venv_python,
            "setup.py",
            "sdist",
            "--dist-dir",
            str(cache_dir),
            cwd=src_dir,
        )
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        salt_generated_package = list(cache_dir.glob("*.tar.gz"))
        if not salt_generated_package:
            pytest.fail("Could not find the generated sdist file")
        salt_generated_package = salt_generated_package[0]
        log.info("Generated sdist file: %s", salt_generated_package.name)

        # Assert generated sdist version matches what salt reports as its version
        sdist_ver_cmp = salt_generated_package.name.split(".tar.gz")[0].split("salt-")[
            -1
        ]
        assert sdist_ver_cmp == salt.version.__version__, "{} != {}".format(
            sdist_ver_cmp, salt.version.__version__
        )

        venv.install(str(salt_generated_package))

        # Let's also ensure we have a salt/_version.py from the installed salt wheel
        subdir = [
            "lib",
            "python{}.{}".format(*sys.version_info),
            "site-packages",
            "salt",
        ]
        if salt.utils.platform.is_windows():
            subdir.pop(1)

        installed_salt_path = pathlib.Path(venv.venv_dir)
        installed_salt_path = installed_salt_path.joinpath(*subdir)
        assert installed_salt_path.is_dir()
        salt_generated_version_file_path = installed_salt_path / "_version.py"
        assert salt_generated_version_file_path.is_file()
        with salt_generated_version_file_path.open() as rfh:
            log.debug("_version.py contents:\n >>>>>>\n%s\n <<<<<<", rfh.read())

        # Let's ensure the version is correct
        cmd = venv.run(venv.venv_python, "-m", "pip", "list", "--format", "json")
        for details in json.loads(cmd.stdout):
            if details["name"] != "salt":
                continue
            installed_version = details["version"]
            break
        else:
            pytest.fail("Salt was not found installed")

        # Let's compare the installed version with the version salt reports
        assert installed_version == salt.version.__version__, "{} != {}".format(
            installed_version, salt.version.__version__
        )


def test_setup_install(virtualenv, cache_dir, use_static_requirements, src_dir):
    """
    test installing directly from source
    """
    # Let's create the testing virtualenv
    with virtualenv as venv:
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        # Setuptools installs pre-release packages if we don't pin to an exact version
        # Let's download and install requirements before, running salt's install test
        venv.run(
            venv.venv_python,
            "-m",
            "pip",
            "download",
            "--dest",
            str(cache_dir),
            src_dir,
        )
        packages = []
        for fname in cache_dir.iterdir():
            if (
                fname.name.startswith("pycurl")
                and salt.utils.platform.is_windows()
                and not use_static_requirements
            ):
                # On windows, the latest pycurl release, 7.43.0.6 at the time of writing,
                # does not have wheel files uploaded, so, delete the downloaded source
                # tarball and will later force pycurl==7.43.0.5 to be pre-installed before
                # installing salt
                fname.unlink()
                continue
            packages.append(fname)
        venv.install(*[str(pkg) for pkg in packages])
        for package in packages:
            package.unlink()

        if salt.utils.platform.is_windows() and not use_static_requirements:
            # Like mentioned above, install pycurl==7.43.0.5
            # However, on windows, the latest pycurl release, 7.43.0.6 at the time of writing,
            # does not have wheel files uploaded, so, we force pycurl==7.43.0.5 to be
            # pre-installed before installing salt
            venv.install("pycurl==7.43.0.5")

        venv.run(
            venv.venv_python,
            "setup.py",
            "install",
            "--prefix",
            str(venv.venv_dir),
            cwd=src_dir,
        )

        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        # Let's ensure the version is correct
        cmd = venv.run(venv.venv_python, "-m", "pip", "list", "--format", "json")
        for details in json.loads(cmd.stdout):
            if details["name"] != "salt":
                continue
            installed_version = details["version"]
            break
        else:
            pytest.fail("Salt was not found installed")

        # Let's compare the installed version with the version salt reports
        assert installed_version == salt.version.__version__, "{} != {}".format(
            installed_version, salt.version.__version__
        )

        # Let's also ensure we have a salt/_version.py from the installed salt
        subdir = [
            "lib",
            "python{}.{}".format(*sys.version_info),
            "site-packages",
        ]
        if salt.utils.platform.is_windows():
            subdir.pop(1)
        site_packages_dir = pathlib.Path(venv.venv_dir)
        site_packages_dir = site_packages_dir.joinpath(*subdir)
        assert site_packages_dir.is_dir()
        installed_salt_path = list(site_packages_dir.glob("salt*.egg"))
        if not installed_salt_path:
            pytest.fail("Failed to find the installed salt path")
        installed_salt_path = installed_salt_path[0] / "salt"
        salt_generated_version_file_path = installed_salt_path / "_version.py"
        assert salt_generated_version_file_path.is_file()



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.mark.slow_test
def test_sync_grains(salt_call_cli):
    ret = salt_call_cli.run("saltutil.sync_grains")
    assert ret.exitcode == 0
    assert ret.json == []



"""
Tests for various minion timeouts
"""


import os
import sys

import pytest
import salt.utils.platform
from tests.support.case import ShellCase


@pytest.mark.windows_whitelisted
class MinionTimeoutTestCase(ShellCase):
    """
    Test minion timing functions
    """

    @pytest.mark.slow_test
    def test_long_running_job(self):
        """
        Test that we will wait longer than the job timeout for a minion to
        return.
        """
        # Launch the command
        sleep_length = 30
        if salt.utils.platform.is_windows():
            popen_kwargs = {"env": dict(os.environ, PYTHONPATH=";".join(sys.path))}
        else:
            popen_kwargs = None
        ret = self.run_salt(
            "minion test.sleep {}".format(sleep_length),
            timeout=90,
            catch_stderr=True,
            popen_kwargs=popen_kwargs,
        )
        self.assertTrue(
            isinstance(ret[0], list),
            "Return is not a list. Minion may have returned error: {}".format(ret),
        )
        self.assertEqual(len(ret[0]), 2, "Standard out wrong length {}".format(ret))
        self.assertTrue(
            "True" in ret[0][1],
            "Minion did not return True after {} seconds. ret={}".format(
                sleep_length, ret
            ),
        )



import salt.pillar.pepa as pepa
from tests.support.unit import TestCase

try:
    from salt.utils.odict import OrderedDict
except ImportError:
    from collections import OrderedDict


class PepaPillarTestCase(TestCase):
    def test_repeated_keys(self):
        expected_result = {
            "foo": {
                "bar": {
                    "foo": True,
                    "baz": True,
                },
            },
        }
        data = OrderedDict(
            [
                ("foo..bar..foo", True),
                ("foo..bar..baz", True),
            ]
        )
        result = pepa.key_value_to_tree(data)
        self.assertDictEqual(result, expected_result)



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""


import salt.modules.launchctl_service as launchctl
import salt.utils.stringutils
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class LaunchctlTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.launchctl
    """

    def setup_loader_modules(self):
        return {launchctl: {}}

    def test_get_all(self):
        """
        Test for Return all installed services
        """
        with patch.dict(
            launchctl.__salt__, {"cmd.run": MagicMock(return_value="A\tB\tC\t\n")}
        ):
            with patch.object(
                launchctl, "_available_services", return_value={"A": "a", "B": "b"}
            ):
                self.assertEqual(launchctl.get_all(), ["A", "B", "C"])

    def test_available(self):
        """
        Test for Check that the given service is available.
        """
        with patch.object(launchctl, "_service_by_name", return_value=True):
            self.assertTrue(launchctl.available("job_label"))

    def test_missing(self):
        """
        Test for The inverse of service.available
        """
        with patch.object(launchctl, "_service_by_name", return_value=True):
            self.assertFalse(launchctl.missing("job_label"))

    def test_status(self):
        """
        Test for Return the status for a service
        """
        launchctl_data = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
    <key>Label</key>
    <string>salt-minion</string>
    <key>LastExitStatus</key>
    <integer>0</integer>
    <key>LimitLoadToSessionType</key>
    <string>System</string>
    <key>OnDemand</key>
    <false/>
    <key>PID</key>
    <integer>71</integer>
    <key>ProgramArguments</key>
    <array>
        <string>/usr/local/bin/salt-minion</string>
    </array>
    <key>TimeOut</key>
    <integer>30</integer>
</dict>
</plist>"""
        with patch.object(
            launchctl, "_service_by_name", return_value={"plist": {"Label": "A"}}
        ):
            launchctl_data = salt.utils.stringutils.to_bytes(launchctl_data)
            with patch.object(
                launchctl, "_get_launchctl_data", return_value=launchctl_data
            ):
                self.assertTrue(launchctl.status("job_label"))

    def test_stop(self):
        """
        Test for Stop the specified service
        """
        with patch.object(
            launchctl, "_service_by_name", return_value={"file_path": "A"}
        ):
            with patch.dict(
                launchctl.__salt__, {"cmd.retcode": MagicMock(return_value=False)}
            ):
                self.assertTrue(launchctl.stop("job_label"))

        with patch.object(launchctl, "_service_by_name", return_value=None):
            self.assertFalse(launchctl.stop("job_label"))

    def test_start(self):
        """
        Test for Start the specified service
        """
        with patch.object(
            launchctl, "_service_by_name", return_value={"file_path": "A"}
        ):
            with patch.dict(
                launchctl.__salt__, {"cmd.retcode": MagicMock(return_value=False)}
            ):
                self.assertTrue(launchctl.start("job_label"))

        with patch.object(launchctl, "_service_by_name", return_value=None):
            self.assertFalse(launchctl.start("job_label"))

    def test_restart(self):
        """
        Test for Restart the named service
        """
        with patch.object(launchctl, "stop", return_value=None):
            with patch.object(launchctl, "start", return_value=True):
                self.assertTrue(launchctl.restart("job_label"))



"""
NAPALM syslog engine
====================

.. versionadded:: 2017.7.0

An engine that takes syslog messages structured in
OpenConfig_ or IETF format
and fires Salt events.

.. _OpenConfig: http://www.openconfig.net/

As there can be many messages pushed into the event bus,
the user is able to filter based on the object structure.

Requirements
------------

- `napalm-logs`_

.. _`napalm-logs`: https://github.com/napalm-automation/napalm-logs

This engine transfers objects from the napalm-logs library
into the event bus. The top dictionary has the following keys:

- ``ip``
- ``host``
- ``timestamp``
- ``os``: the network OS identified
- ``model_name``: the OpenConfig or IETF model name
- ``error``: the error name (consult the documentation)
- ``message_details``: details extracted from the syslog message
- ``open_config``: the OpenConfig model

The napalm-logs transfers the messages via widely used transport
mechanisms such as: ZeroMQ (default), Kafka, etc.

The user can select the right transport using the ``transport``
option in the configuration.

:configuration: Example configuration

    .. code-block:: yaml

        engines:
          - napalm_syslog:
              transport: zmq
              address: 1.2.3.4
              port: 49018

:configuration: Configuration example, excluding messages from IOS-XR devices:

    .. code-block:: yaml

        engines:
          - napalm_syslog:
              transport: kafka
              address: 1.2.3.4
              port: 49018
              os_blacklist:
                - iosxr

Event example:

.. code-block:: json

    {
        "_stamp": "2017-05-26T10:03:18.653045",
        "error": "BGP_PREFIX_THRESH_EXCEEDED",
        "host": "vmx01",
        "ip": "192.168.140.252",
        "message_details": {
            "date": "May 25",
            "host": "vmx01",
            "message": "192.168.140.254 (External AS 65001): Configured maximum prefix-limit threshold(22) exceeded for inet-unicast nlri: 28 (instance master)",
            "pri": "28",
            "processId": "2957",
            "processName": "rpd",
            "tag": "BGP_PREFIX_THRESH_EXCEEDED",
            "time": "20:50:41"
        },
        "model_name": "openconfig_bgp",
        "open_config": {
            "bgp": {
                "neighbors": {
                    "neighbor": {
                        "192.168.140.254": {
                            "afi_safis": {
                                "afi_safi": {
                                    "inet": {
                                        "afi_safi_name": "inet",
                                        "ipv4_unicast": {
                                            "prefix_limit": {
                                                "state": {
                                                    "max_prefixes": 22
                                                }
                                            }
                                        },
                                        "state": {
                                            "prefixes": {
                                                "received": 28
                                            }
                                        }
                                    }
                                }
                            },
                            "neighbor_address": "192.168.140.254",
                            "state": {
                                "peer_as": 65001
                            }
                        }
                    }
                }
            }
        },
        "os": "junos",
        "timestamp": "1495741841"
    }

To consume the events and eventually react and deploy a configuration changes
on the device(s) firing the event, one is able to identify the minion ID, using
one of the following alternatives, but not limited to:

- :mod:`Host grains <salt.grains.napalm.host>` to match the event tag
- :mod:`Host DNS grain <salt.grains.napalm.host_dns>` to match the IP address in the event data
- :mod:`Hostname grains <salt.grains.napalm.hostname>` to match the event tag
- :ref:`Define static grains <static-custom-grains>`
- :ref:`Write a grains module <writing-grains>`
- :ref:`Targeting minions using pillar data <targeting-pillar>` - The user can
  configure certain information in the Pillar data and then use it to identify
  minions

Master configuration example, to match the event and react:

.. code-block:: yaml

    reactor:
      - 'napalm/syslog/*/BGP_PREFIX_THRESH_EXCEEDED/*':
        - salt://increase_prefix_limit_on_thresh_exceeded.sls

Which matches the events having the error code ``BGP_PREFIX_THRESH_EXCEEDED``
from any network operating system, from any host and reacts, executing the
``increase_prefix_limit_on_thresh_exceeded.sls`` reactor, found under
one of the :conf_master:`file_roots` paths.

Reactor example:

.. code-block:: yaml

    increase_prefix_limit_on_thresh_exceeded:
      local.net.load_template:
        - tgt: "hostname:{{ data['host'] }}"
        - tgt_type: grain
        - kwarg:
            template_name: salt://increase_prefix_limit.jinja
            openconfig_structure: {{ data['open_config'] }}

The reactor in the example increases the BGP prefix limit
when triggered by an event as above. The minion is matched using the ``host``
field from the ``data`` (which is the body of the event), compared to the
:mod:`hostname grain <salt.grains.napalm.hostname>` field. When the event
occurs, the reactor will execute the
:mod:`net.load_template <salt.modules.napalm_network.load_template>` function,
sending as arguments the template ``salt://increase_prefix_limit.jinja`` defined
by the user in their environment and the complete OpenConfig object under
the variable name ``openconfig_structure``. Inside the Jinja template, the user
can process the object from ``openconfig_structure`` and define the bussiness
logic as required.
"""

import logging

import salt.utils.event as event
import salt.utils.network
import salt.utils.stringutils
from salt.utils.zeromq import zmq

try:
    # pylint: disable=import-error
    import napalm_logs
    import napalm_logs.utils

    # pylint: enable=import-error
    HAS_NAPALM_LOGS = True
except ImportError:
    HAS_NAPALM_LOGS = False


log = logging.getLogger(__name__)

__virtualname__ = "napalm_syslog"


def __virtual__():
    """
    Load only if napalm-logs is installed.
    """
    if not HAS_NAPALM_LOGS or not zmq:
        return (
            False,
            "napalm_syslog could not be loaded. Please install "
            "napalm-logs library and ZeroMQ.",
        )
    return True


def _zmq(address, port, **kwargs):
    context = zmq.Context()
    socket = context.socket(zmq.SUB)
    if salt.utils.network.is_ipv6(address):
        socket.ipv6 = True
    socket.connect("tcp://{addr}:{port}".format(addr=address, port=port))
    socket.setsockopt(zmq.SUBSCRIBE, b"")
    return socket.recv


def _get_transport_recv(name="zmq", address="0.0.0.0", port=49017, **kwargs):
    if name not in TRANSPORT_FUN_MAP:
        log.error("Invalid transport: %s. Falling back to ZeroMQ.", name)
        name = "zmq"
    return TRANSPORT_FUN_MAP[name](address, port, **kwargs)


TRANSPORT_FUN_MAP = {"zmq": _zmq, "zeromq": _zmq}


def start(
    transport="zmq",
    address="0.0.0.0",
    port=49017,
    auth_address="0.0.0.0",
    auth_port=49018,
    disable_security=False,
    certificate=None,
    os_whitelist=None,
    os_blacklist=None,
    error_whitelist=None,
    error_blacklist=None,
    host_whitelist=None,
    host_blacklist=None,
):
    """
    Listen to napalm-logs and publish events into the Salt event bus.

    transport: ``zmq``
        Choose the desired transport.

        .. note::
            Currently ``zmq`` is the only valid option.

    address: ``0.0.0.0``
        The address of the publisher, as configured on napalm-logs.

    port: ``49017``
        The port of the publisher, as configured on napalm-logs.

    auth_address: ``0.0.0.0``
        The address used for authentication
        when security is not disabled.

    auth_port: ``49018``
        Port used for authentication.

    disable_security: ``False``
        Trust unencrypted messages.
        Strongly discouraged in production.

    certificate: ``None``
        Absolute path to the SSL certificate.

    os_whitelist: ``None``
        List of operating systems allowed. By default everything is allowed.

    os_blacklist: ``None``
        List of operating system to be ignored. Nothing ignored by default.

    error_whitelist: ``None``
        List of errors allowed.

    error_blacklist: ``None``
        List of errors ignored.

    host_whitelist: ``None``
        List of hosts or IPs to be allowed.

    host_blacklist: ``None``
        List of hosts of IPs to be ignored.
    """
    if not disable_security:
        if not certificate:
            log.critical("Please use a certificate, or disable the security.")
            return
        auth = napalm_logs.utils.ClientAuth(
            certificate, address=auth_address, port=auth_port
        )

    transport_recv_fun = _get_transport_recv(name=transport, address=address, port=port)
    if not transport_recv_fun:
        log.critical("Unable to start the engine", exc_info=True)
        return
    master = False
    if __opts__["__role"] == "master":
        master = True
    while True:
        log.debug("Waiting for napalm-logs to send anything...")
        raw_object = transport_recv_fun()
        log.debug("Received from napalm-logs:")
        log.debug(raw_object)
        if not disable_security:
            dict_object = auth.decrypt(raw_object)
        else:
            dict_object = napalm_logs.utils.unserialize(raw_object)
        try:
            event_os = dict_object["os"]
            if os_blacklist or os_whitelist:
                valid_os = salt.utils.stringutils.check_whitelist_blacklist(
                    event_os, whitelist=os_whitelist, blacklist=os_blacklist
                )
                if not valid_os:
                    log.info("Ignoring NOS %s as per whitelist/blacklist", event_os)
                    continue
            event_error = dict_object["error"]
            if error_blacklist or error_whitelist:
                valid_error = salt.utils.stringutils.check_whitelist_blacklist(
                    event_error, whitelist=error_whitelist, blacklist=error_blacklist
                )
                if not valid_error:
                    log.info(
                        "Ignoring error %s as per whitelist/blacklist", event_error
                    )
                    continue
            event_host = dict_object.get("host") or dict_object.get("ip")
            if host_blacklist or host_whitelist:
                valid_host = salt.utils.stringutils.check_whitelist_blacklist(
                    event_host, whitelist=host_whitelist, blacklist=host_blacklist
                )
                if not valid_host:
                    log.info(
                        "Ignoring messages from %s as per whitelist/blacklist",
                        event_host,
                    )
                    continue
            tag = "napalm/syslog/{os}/{error}/{host}".format(
                os=event_os, error=event_error, host=event_host
            )
        except KeyError as kerr:
            log.warning("Missing keys from the napalm-logs object:", exc_info=True)
            log.warning(dict_object)
            continue  # jump to the next object in the queue
        log.debug("Sending event %s", tag)
        log.debug(raw_object)
        if master:
            event.get_master_event(__opts__, __opts__["sock_dir"]).fire_event(
                dict_object, tag
            )
        else:
            __salt__["event.send"](tag, dict_object)



r"""
A salt util for modifying the audit policies on the machine. This util is used
by the ``win_auditpol`` and ``win_lgpo`` modules.

Though this utility does not set group policy for auditing, it displays how all
auditing configuration is applied on the machine, either set directly or via
local or domain group policy.

.. versionadded:: 2018.3.4
.. versionadded:: 2019.2.1

This util allows you to view and modify the audit settings as they are applied
on the machine. The audit settings are broken down into nine categories:

- Account Logon
- Account Management
- Detailed Tracking
- DS Access
- Logon/Logoff
- Object Access
- Policy Change
- Privilege Use
- System

The ``get_settings`` function will return the subcategories for all nine of
the above categories in one dictionary along with their auditing status.

To modify a setting you only need to specify the subcategory name and the value
you wish to set. Valid settings are:

- No Auditing
- Success
- Failure
- Success and Failure

Usage:

.. code-block:: python

    import salt.utils.win_lgpo_auditpol

    # Get current state of all audit settings
    salt.utils.win_lgpo_auditpol.get_settings()

    # Get the current state of all audit settings in the "Account Logon"
    # category
    salt.utils.win_lgpo_auditpol.get_settings(category="Account Logon")

    # Get current state of the "Credential Validation" setting
    salt.utils.win_lgpo_auditpol.get_setting(name='Credential Validation')

    # Set the state of the "Credential Validation" setting to Success and
    # Failure
    salt.utils.win_lgpo_auditpol.set_setting(name='Credential Validation',
                                             value='Success and Failure')

    # Set the state of the "Credential Validation" setting to No Auditing
    salt.utils.win_lgpo_auditpol.set_setting(name='Credential Validation',
                                             value='No Auditing')
"""

import logging
import re
import tempfile

import salt.modules.cmdmod
import salt.utils.files
import salt.utils.platform
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)
__virtualname__ = "auditpol"

categories = [
    "Account Logon",
    "Account Management",
    "Detailed Tracking",
    "DS Access",
    "Logon/Logoff",
    "Object Access",
    "Policy Change",
    "Privilege Use",
    "System",
]

settings = {
    "No Auditing": "/success:disable /failure:disable",
    "Success": "/success:enable /failure:disable",
    "Failure": "/success:disable /failure:enable",
    "Success and Failure": "/success:enable /failure:enable",
}


# Although utils are often directly imported, it is also possible to use the
# loader.
def __virtual__():
    """
    Only load if on a Windows system
    """
    if not salt.utils.platform.is_windows():
        return False, "This utility only available on Windows"

    return __virtualname__


def _auditpol_cmd(cmd):
    """
    Helper function for running the auditpol command

    Args:
        cmd (str): the auditpol command to run

    Returns:
        list: A list containing each line of the return (splitlines)

    Raises:
        CommandExecutionError: If the command encounters an error
    """
    ret = salt.modules.cmdmod.run_all(cmd="auditpol {}".format(cmd), python_shell=True)
    if ret["retcode"] == 0:
        return ret["stdout"].splitlines()

    msg = "Error executing auditpol command: {}\n".format(cmd)
    msg += "\n".join(ret["stdout"])
    raise CommandExecutionError(msg)


def get_settings(category="All"):
    """
    Get the current configuration for all audit settings specified in the
    category

    Args:
        category (str):
            One of the nine categories to return. Can also be ``All`` to return
            the settings for all categories. Valid options are:

            - Account Logon
            - Account Management
            - Detailed Tracking
            - DS Access
            - Logon/Logoff
            - Object Access
            - Policy Change
            - Privilege Use
            - System
            - All

            Default value is ``All``

    Returns:
        dict: A dictionary containing all subcategories for the specified
            category along with their current configuration

    Raises:
        KeyError: On invalid category
        CommandExecutionError: If an error is encountered retrieving the settings

    Usage:

    .. code-block:: python

        import salt.utils.win_lgpo_auditpol

        # Get current state of all audit settings
        salt.utils.win_lgpo_auditpol.get_settings()

        # Get the current state of all audit settings in the "Account Logon"
        # category
        salt.utils.win_lgpo_auditpol.get_settings(category="Account Logon")
    """
    # Parameter validation
    if category.lower() in ["all", "*"]:
        category = "*"
    elif category.lower() not in [x.lower() for x in categories]:
        raise KeyError('Invalid category: "{}"'.format(category))

    cmd = '/get /category:"{}"'.format(category)
    results = _auditpol_cmd(cmd)

    ret = {}
    # Skip the first 2 lines
    for line in results[3:]:
        if "  " in line.strip():
            ret.update(dict(list(zip(*[iter(re.split(r"\s{2,}", line.strip()))] * 2))))
    return ret


def get_setting(name):
    """
    Get the current configuration for the named audit setting

    Args:
        name (str): The name of the setting to retrieve

    Returns:
        str: The current configuration for the named setting

    Raises:
        KeyError: On invalid setting name
        CommandExecutionError: If an error is encountered retrieving the settings

    Usage:

    .. code-block:: python

        import salt.utils.win_lgpo_auditpol

        # Get current state of the "Credential Validation" setting
        salt.utils.win_lgpo_auditpol.get_setting(name='Credential Validation')
    """
    current_settings = get_settings(category="All")
    for setting in current_settings:
        if name.lower() == setting.lower():
            return current_settings[setting]
    raise KeyError("Invalid name: {}".format(name))


def _get_valid_names():
    if "auditpol.valid_names" not in __context__:
        settings = get_settings(category="All")
        __context__["auditpol.valid_names"] = [k.lower() for k in settings]
    return __context__["auditpol.valid_names"]


def set_setting(name, value):
    """
    Set the configuration for the named audit setting

    Args:

        name (str):
            The name of the setting to configure

        value (str):
            The configuration for the named value. Valid options are:

            - No Auditing
            - Success
            - Failure
            - Success and Failure

    Returns:
        bool: True if successful

    Raises:
        KeyError: On invalid ``name`` or ``value``
        CommandExecutionError: If an error is encountered modifying the setting

    Usage:

    .. code-block:: python

        import salt.utils.win_lgpo_auditpol

        # Set the state of the "Credential Validation" setting to Success and
        # Failure
        salt.utils.win_lgpo_auditpol.set_setting(name='Credential Validation',
                                                 value='Success and Failure')

        # Set the state of the "Credential Validation" setting to No Auditing
        salt.utils.win_lgpo_auditpol.set_setting(name='Credential Validation',
                                                 value='No Auditing')
    """
    # Input validation
    if name.lower() not in _get_valid_names():
        raise KeyError("Invalid name: {}".format(name))
    for setting in settings:
        if value.lower() == setting.lower():
            cmd = '/set /subcategory:"{}" {}'.format(name, settings[setting])
            break
    else:
        raise KeyError("Invalid setting value: {}".format(value))

    _auditpol_cmd(cmd)

    return True


def get_auditpol_dump():
    """
    Gets the contents of an auditpol /backup. Used by the LGPO module to get
    fieldnames and GUIDs for Advanced Audit policies.

    Returns:
        list: A list of lines form the backup file

    Usage:

    .. code-block:: python

        import salt.utils.win_lgpo_auditpol

        dump = salt.utils.win_lgpo_auditpol.get_auditpol_dump()
    """
    # Just get a temporary file name
    # NamedTemporaryFile will delete the file it creates by default on Windows
    with tempfile.NamedTemporaryFile(suffix=".csv") as tmp_file:
        csv_file = tmp_file.name

    cmd = "/backup /file:{}".format(csv_file)
    _auditpol_cmd(cmd)

    with salt.utils.files.fopen(csv_file) as fp:
        return fp.readlines()



import os

import pytest
import salt.utils.files
import salt.utils.platform
from tests.support.case import ModuleCase
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import skipIf


@skipIf(not salt.utils.platform.is_windows(), "Tests for only Windows")
class DscModuleTest(ModuleCase):
    """
    Validate PowerShell DSC module
    """

    def setUp(self):
        self.ps1file = os.path.join(RUNTIME_VARS.TMP, "HelloWorld.ps1")
        with salt.utils.files.fopen(
            os.path.join(RUNTIME_VARS.FILES, "file", "base", "HelloWorld.ps1"), "rb"
        ) as sfp:
            with salt.utils.files.fopen(self.ps1file, "wb") as dfp:
                dfp.write(sfp.read())
        self.psd1file = os.path.join(RUNTIME_VARS.TMP, "HelloWorld.psd1")
        with salt.utils.files.fopen(
            os.path.join(RUNTIME_VARS.FILES, "file", "base", "HelloWorld.psd1"), "rb"
        ) as sfp:
            with salt.utils.files.fopen(self.psd1file, "wb") as dfp:
                dfp.write(sfp.read())
        super().setUp()

    def tearDown(self):
        if os.path.isfile(self.ps1file):
            os.remove(self.ps1file)
        if os.path.isfile(self.psd1file):
            os.remove(self.psd1file)
        super().tearDown()

    @pytest.mark.destructive_test
    def test_compile_config(self):
        ret = self.run_function(
            "dsc.compile_config",
            self.ps1file,
            config_name="HelloWorld",
            config_data=self.psd1file,
        )
        self.assertTrue(ret["Exists"])



"""
The Salt Cloud Runner
=====================

This runner wraps the functionality of salt cloud making salt cloud routines
available to all internal apis via the runner system
"""

import logging
import os

import salt.cloud
import salt.utils.args
from salt.exceptions import SaltCloudConfigError

# Get logging started
log = logging.getLogger(__name__)


def _get_client():
    """
    Return cloud client
    """
    client = salt.cloud.CloudClient(
        os.path.join(os.path.dirname(__opts__["conf_file"]), "cloud")
    )
    return client


def list_sizes(provider="all"):
    """
    List cloud provider sizes for the given providers
    """
    client = _get_client()
    sizes = client.list_sizes(provider)
    return sizes


def list_images(provider="all"):
    """
    List cloud provider images for the given providers
    """
    client = _get_client()
    images = client.list_images(provider)
    return images


def list_locations(provider="all"):
    """
    List cloud provider sizes for the given providers
    """
    client = _get_client()
    locations = client.list_locations(provider)
    return locations


def query(query_type="list_nodes"):
    """
    List cloud provider data for all providers
    """
    client = _get_client()
    info = client.query(query_type)
    return info


def full_query(query_type="list_nodes_full"):
    """
    List all available cloud provider data
    """
    client = _get_client()
    info = client.full_query(query_type)
    return info


def select_query(query_type="list_nodes_select"):
    """
    List selected nodes
    """
    client = _get_client()
    info = client.select_query(query_type)
    return info


def profile(prof=None, instances=None, opts=None, **kwargs):
    """
    Create a cloud vm with the given profile and instances, instances can be a
    list or comma-delimited string

    CLI Example:

    .. code-block:: bash

        salt-run cloud.profile prof=my-ec2 instances=node1,node2,node3
    """
    if prof is None and "profile" in kwargs:
        prof = kwargs["profile"]

    if prof is None:
        return {"Error": "A profile (or prof) must be defined"}

    if instances is None and "names" in kwargs:
        instances = kwargs["names"]

    if instances is None:
        return {"Error": "One or more instances (comma-delimited) must be set"}

    client = _get_client()
    if isinstance(opts, dict):
        client.opts.update(opts)
    info = client.profile(prof, instances, **salt.utils.args.clean_kwargs(**kwargs))
    return info


def map_run(path=None, opts=None, **kwargs):
    """
    Execute a salt cloud map file
    """
    client = _get_client()
    if isinstance(opts, dict):
        client.opts.update(opts)
    info = client.map_run(path, **salt.utils.args.clean_kwargs(**kwargs))
    return info


def destroy(instances, opts=None):
    """
    Destroy the named vm(s)
    """
    client = _get_client()
    if isinstance(opts, dict):
        client.opts.update(opts)
    info = client.destroy(instances)
    return info


def action(
    func=None,
    cloudmap=None,
    instances=None,
    provider=None,
    instance=None,
    opts=None,
    **kwargs
):
    """
    Execute a single action on the given map/provider/instance

    CLI Example:

    .. code-block:: bash

        salt-run cloud.action start my-salt-vm
    """
    info = {}
    client = _get_client()
    if isinstance(opts, dict):
        client.opts.update(opts)
    try:
        info = client.action(
            func,
            cloudmap,
            instances,
            provider,
            instance,
            salt.utils.args.clean_kwargs(**kwargs),
        )
    except SaltCloudConfigError as err:
        log.error(err)
    return info


def create(provider, instances, opts=None, **kwargs):
    """
    Create an instance using Salt Cloud

    CLI Example:

    .. code-block:: bash

        salt-run cloud.create my-ec2-config myinstance \\
            image=ami-1624987f size='t1.micro' ssh_username=ec2-user \\
            securitygroup=default delvol_on_destroy=True
    """
    client = _get_client()
    if isinstance(opts, dict):
        client.opts.update(opts)
    info = client.create(provider, instances, **salt.utils.args.clean_kwargs(**kwargs))
    return info
